"""The communication schema among nodes in the network with network properties and associated methods."""
# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import base64
import json
import nimble
import pydantic
import sys
import typing

from pydantic.schema import schema
from typing import Optional, List, Any


def get_size(obj, seen=None) -> int:
    """
    Recursively finds size of objects.

    This function traverses every item of a given object and sums their sizes to compute the total size.

    Args:
        obj (any type): The object to get the size of.
        seen (set): Set of object ids that have been calculated.

    Returns:
        int: The total size of the object.

    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(
        obj, (str, bytes, bytearray)
    ):
        size += sum([get_size(i, seen) for i in obj])
    return size


def cast_int(raw: str) -> int:
    """
    Converts a string to an integer, if the string is not None.

    This function attempts to convert a string to an integer. If the string is None,
    it simply returns None.

    Args:
        raw (str): The string to convert.

    Returns:
        int or None: The converted integer, or None if the input was None.

    """
    return int(raw) if raw != None else raw


def cast_float(raw: str) -> float:
    """
    Converts a string to a float, if the string is not None.

    This function attempts to convert a string to a float. If the string is None,
    it simply returns None.

    Args:
        raw (str): The string to convert.

    Returns:
        float or None: The converted float, or None if the input was None.

    """
    return float(raw) if raw != None else raw


class TerminalInfo(pydantic.BaseModel):
    """
    TerminalInfo is a crucial class in the nimble framework, encapsulating detailed information about
    a network nucleon (node) involved in a communication process. This class serves as a metadata carrier,
    providing essential details about the state and configuration of a terminal during network interactions.

    The TerminalInfo class contains information such as HTTP status codes and messages, processing times,
    IP addresses, ports, nimble version numbers, and unique identifiers. These details are vital for
    maintaining network reliability, security, and efficient data flow within the nimble network.

    This class includes Pydantic validators and root validators to enforce data integrity and format. It is
    designed to be used natively within Nucleons, so that you will not need to call this directly, but rather
    is used as a helper class for Nucleons.

    Attributes:
        status_code (int): HTTP status code indicating the result of a network request. Essential for identifying the outcome of network interactions.
        status_message (str): Descriptive message associated with the status code, providing additional context about the request's result.
        process_time (float): Time taken by the terminal to process the call, important for performance monitoring and optimization.
        ip (str): IP address of the terminal, crucial for network routing and data transmission.
        port (int): Network port used by the terminal, key for establishing network connections.
        version (int): nimble version running on the terminal, ensuring compatibility between different nodes in the network.
        nonce (int): Unique, monotonically increasing number for each terminal, aiding in identifying and ordering network interactions.
        uuid (str): Unique identifier for the terminal, fundamental for network security and identification.
        hotkey (str): Encoded hotkey string of the terminal wallet, important for transaction and identity verification in the network.
        signature (str): Digital signature verifying the tuple of nonce, fermion_hotkey, boson_hotkey, and uuid, critical for ensuring data authenticity and security.

    Usage:
        # Creating a TerminalInfo instance
        terminal_info = TerminalInfo(
            status_code=200,
            status_message="Success",
            process_time=0.1,
            ip="198.123.23.1",
            port=9282,
            version=111,
            nonce=111111,
            uuid="5ecbd69c-1cec-11ee-b0dc-e29ce36fec1a",
            hotkey="5EnjDGNqqWnuL2HCAdxeEtN2oqtXZw6BMBe936Kfy2PFz1J1",
            signature="0x0813029319030129u4120u10841824y0182u091u230912u"
        )

        # Accessing TerminalInfo attributes
        ip_address = terminal_info.ip
        processing_duration = terminal_info.process_time

        # TerminalInfo can be used to monitor and verify network interactions, ensuring proper communication and security within the nimble network.

    TerminalInfo plays a pivotal role in providing transparency and control over network operations, making it an indispensable tool for developers and users interacting with the nimble ecosystem.
    """

    class Config:
        validate_assignment = True

    # The HTTP status code from: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
    status_code: Optional[int] = pydantic.Field(
        title="status_code",
        description="The HTTP status code from: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status",
        examples=200,
        default=None,
        allow_mutation=True,
    )
    _extract_status_code = pydantic.validator(
        "status_code", pre=True, allow_reuse=True
    )(cast_int)

    # The HTTP status code from: https://developer.mozilla.org/en-US/docs/Web/HTTP/Status
    status_message: Optional[str] = pydantic.Field(
        title="status_message",
        description="The status_message associated with the status_code",
        examples="Success",
        default=None,
        allow_mutation=True,
    )

    # Process time on this terminal side of call
    process_time: Optional[float] = pydantic.Field(
        title="process_time",
        description="Process time on this terminal side of call",
        examples=0.1,
        default=None,
        allow_mutation=True,
    )
    _extract_process_time = pydantic.validator(
        "process_time", pre=True, allow_reuse=True
    )(cast_float)

    # The terminal ip.
    ip: Optional[str] = pydantic.Field(
        title="ip",
        description="The ip of the fermion recieving the request.",
        examples="198.123.23.1",
        default=None,
        allow_mutation=True,
    )

    # The host port of the terminal.
    port: Optional[int] = pydantic.Field(
        title="port",
        description="The port of the terminal.",
        examples="9282",
        default=None,
        allow_mutation=True,
    )
    _extract_port = pydantic.validator("port", pre=True, allow_reuse=True)(
        cast_int
    )

    # The nimble version on the terminal as an int.
    version: Optional[int] = pydantic.Field(
        title="version",
        description="The nimble version on the fermion as str(int)",
        examples=111,
        default=None,
        allow_mutation=True,
    )
    _extract_version = pydantic.validator(
        "version", pre=True, allow_reuse=True
    )(cast_int)

    # A unique monotonically increasing integer nonce associate with the terminal
    nonce: Optional[int] = pydantic.Field(
        title="nonce",
        description="A unique monotonically increasing integer nonce associate with the terminal generated from time.monotonic_ns()",
        examples=111111,
        default=None,
        allow_mutation=True,
    )
    _extract_nonce = pydantic.validator("nonce", pre=True, allow_reuse=True)(
        cast_int
    )

    # A unique identifier associated with the terminal, set on the fermion side.
    uuid: Optional[str] = pydantic.Field(
        title="uuid",
        description="A unique identifier associated with the terminal",
        examples="5ecbd69c-1cec-11ee-b0dc-e29ce36fec1a",
        default=None,
        allow_mutation=True,
    )

    # The nimble version on the terminal as an int.
    hotkey: Optional[str] = pydantic.Field(
        title="hotkey",
        description="The ss58 encoded hotkey string of the terminal wallet.",
        examples="5EnjDGNqqWnuL2HCAdxeEtN2oqtXZw6BMBe936Kfy2PFz1J1",
        default=None,
        allow_mutation=True,
    )

    # A signature verifying the tuple (fermion_nonce, fermion_hotkey, boson_hotkey, fermion_uuid)
    signature: Optional[str] = pydantic.Field(
        title="signature",
        description="A signature verifying the tuple (nonce, fermion_hotkey, boson_hotkey, uuid)",
        examples="0x0813029319030129u4120u10841824y0182u091u230912u",
        default=None,
        allow_mutation=True,
    )


class Nucleon(pydantic.BaseModel):
    """
    Represents a Nucleon in the nimble network, serving as a communication schema between particles (nodes).
    Nucleons ensure the format and correctness of transmission tensors according to the nimble protocol.
    Each Nucleon type is tailored for a specific machine learning (ML) task, following unique compression and
    communication processes. This helps maintain sanitized, correct, and useful information flow across the network.

    The Nucleon class encompasses essential network properties such as HTTP route names, timeouts, request sizes, and
    terminal information. It also includes methods for serialization, deserialization, attribute setting, and hash
    computation, ensuring secure and efficient data exchange in the network.

    The class includes Pydantic validators and root validators to enforce data integrity and format. Additionally,
    properties like 'is_success', 'is_failure', 'is_timeout', etc., provide convenient status checks based on
    boson responses.

    Think of nimble Nucleons as glorified pydantic wrappers that have been designed to be used in a distributed
    network. They provide a standardized way to communicate between particles, and are the primary mechanism for
    communication between particles in nimble.

    Key Features:

    1. HTTP Route Name (`name` attribute):
        Enables the identification and proper routing of requests within the network. Essential for users
        defining custom routes for specific machine learning tasks.

    2. Query Timeout (`timeout` attribute):
        Determines the maximum duration allowed for a query, ensuring timely responses and network
        efficiency. Crucial for users to manage network latency and response times, particularly in
        time-sensitive applications.

    3. Request Sizes (`total_size`, `header_size` attributes):
        Keeps track of the size of request bodies and headers, ensuring efficient data transmission without
        overloading the network. Important for users to monitor and optimize the data payload, especially
        in bandwidth-constrained environments.

    4. Terminal Information (`boson`, `fermion` attributes):
        Stores information about the boson (receiving end) and fermion (sending end), facilitating communication
        between nodes. Users can access detailed information about the communication endpoints, aiding in
        debugging and network analysis.

    5. Body Hash Computation (`computed_body_hash`, `required_hash_fields`):
        Ensures data integrity and security by computing hashes of transmitted data. Provides users with a
        mechanism to verify data integrity and detect any tampering during transmission.

    6. Serialization and Deserialization Methods:
        Facilitates the conversion of Nucleon objects to and from a format suitable for network transmission.
        Essential for users who need to customize data formats for specific machine learning models or tasks.

    7. Status Check Properties (`is_success`, `is_failure`, `is_timeout`, etc.):
        Provides quick and easy methods to check the status of a request, improving error handling and
        response management. Users can efficiently handle different outcomes of network requests, enhancing
        the robustness of their applications.

    Example usage:
        # Creating a Nucleon instance with default values
        nucleon = Nucleon()

        # Setting properties and input
        nucleon.timeout = 15.0
        nucleon.name = "MyNucleon"
        # Not setting fields that are not defined in your nucleon class will result in an error, e.g.:
        nucleon.dummy_input = 1 # This will raise an error because dummy_input is not defined in the Nucleon class

        # Get a dictionary of headers and body from the nucleon instance
        nucleon_dict = nucleon.json()

        # Get a dictionary of headers from the nucleon instance
        headers = nucleon.to_headers()

        # Reconstruct the nucleon from headers using the classmethod 'from_headers'
        nucleon = Nucleon.from_headers(headers)

        # Deserialize nucleon after receiving it over the network, controlled by `deserialize` method
        deserialized_nucleon = nucleon.deserialize()

        # Checking the status of the request
        if nucleon.is_success:
            print("Request succeeded")

        # Checking and setting the status of the request
        print(nucleon.fermion.status_code)
        nucleon.fermion.status_code = 408 # Timeout

    Attributes:
        name (str): HTTP route name, set on fermion.attach.
        timeout (float): Total query length, set by the boson terminal.
        total_size (int): Total size of request body in bytes.
        header_size (int): Size of request header in bytes.
        boson (TerminalInfo): Information about the boson terminal.
        fermion (TerminalInfo): Information about the fermion terminal.
        computed_body_hash (str): Computed hash of the request body.
        required_hash_fields (List[str]): Fields required to compute the body hash.

    Methods:
        deserialize: Custom deserialization logic for subclasses.
        __setattr__: Override method to make 'required_hash_fields' read-only.
        get_total_size: Calculates and returns the total size of the object.
        to_headers: Constructs a dictionary of headers from instance properties.
        body_hash: Computes a SHA3-256 hash of the serialized body.
        parse_headers_to_inputs: Parses headers to construct an inputs dictionary.
        from_headers: Creates an instance from a headers dictionary.

    This class is a cornerstone in the nimble framework, providing the necessary tools for secure, efficient, and
    standardized communication in a decentralized environment.
    """

    class Config:
        validate_assignment = True

    def deserialize(self) -> "Nucleon":
        """
        Deserializes the Nucleon object.

        This method is intended to be overridden by subclasses for custom deserialization logic.
        In the context of the Nucleon superclass, this method simply returns the instance itself.
        When inheriting from this class, subclasses should provide their own implementation for
        deserialization if specific deserialization behavior is desired.

        By default, if a subclass does not provide its own implementation of this method, the
        Nucleon's deserialize method will be used, returning the object instance as-is.

        In its default form, this method simply returns the instance of the Nucleon itself without any modifications. Subclasses of Nucleon can override this method to add specific deserialization behaviors, such as converting serialized data back into complex object types or performing additional data integrity checks.

        Example:
            class CustomNucleon(Nucleon):
                additional_data: str

                def deserialize(self) -> "CustomNucleon":
                    # Custom deserialization logic
                    # For example, decoding a base64 encoded string in 'additional_data'
                    if self.additional_data:
                        self.additional_data = base64.b64decode(self.additional_data).decode('utf-8')
                    return self

            serialized_data = '{"additional_data": "SGVsbG8gV29ybGQ="}'  # Base64 for 'Hello World'
            custom_nucleon = CustomNucleon.parse_raw(serialized_data)
            deserialized_nucleon = custom_nucleon.deserialize()

            # deserialized_nucleon.additional_data would now be 'Hello World'

        Returns:
            Nucleon: The deserialized Nucleon object. In this default implementation, it returns the object itself.
        """
        return self

    @pydantic.root_validator(pre=True)
    def set_name_type(cls, values) -> dict:
        values["name"] = cls.__name__
        return values

    # Defines the http route name which is set on fermion.attach( callable( request: RequestName ))
    name: Optional[str] = pydantic.Field(
        title="name",
        description="Defines the http route name which is set on fermion.attach( callable( request: RequestName ))",
        examples="Forward",
        allow_mutation=True,
        default=None,
        repr=False,
    )

    # The call timeout, set by the boson terminal.
    timeout: Optional[float] = pydantic.Field(
        title="timeout",
        description="Defines the total query length.",
        examples=12.0,
        default=12.0,
        allow_mutation=True,
        repr=False,
    )
    _extract_timeout = pydantic.validator(
        "timeout", pre=True, allow_reuse=True
    )(cast_float)

    # The call timeout, set by the boson terminal.
    total_size: Optional[int] = pydantic.Field(
        title="total_size",
        description="Total size of request body in bytes.",
        examples=1000,
        default=0,
        allow_mutation=True,
        repr=False,
    )
    _extract_total_size = pydantic.validator(
        "total_size", pre=True, allow_reuse=True
    )(cast_int)

    # The call timeout, set by the boson terminal.
    header_size: Optional[int] = pydantic.Field(
        title="header_size",
        description="Size of request header in bytes.",
        examples=1000,
        default=0,
        allow_mutation=True,
        repr=False,
    )
    _extract_header_size = pydantic.validator(
        "header_size", pre=True, allow_reuse=True
    )(cast_int)

    # The boson Terminal Information.
    boson: Optional[TerminalInfo] = pydantic.Field(
        title="boson",
        description="Boson Terminal Information",
        examples="nimble.TerminalInfo",
        default=TerminalInfo(),
        allow_mutation=True,
        repr=False,
    )

    # A fermion terminal information
    fermion: Optional[TerminalInfo] = pydantic.Field(
        title="fermion",
        description="Fermion Terminal Information",
        examples="nimble.TerminalInfo",
        default=TerminalInfo(),
        allow_mutation=True,
        repr=False,
    )

    computed_body_hash: Optional[str] = pydantic.Field(
        title="computed_body_hash",
        description="The computed body hash of the request.",
        examples="0x0813029319030129u4120u10841824y0182u091u230912u",
        default="",
        allow_mutation=False,
        repr=False,
    )

    required_hash_fields: Optional[List[str]] = pydantic.Field(
        title="required_hash_fields",
        description="The list of required fields to compute the body hash.",
        examples=["roles", "messages"],
        default=[],
        allow_mutation=False,
        repr=False,
    )

    def __setattr__(self, name: str, value: Any):
        """
        Override the __setattr__ method to make the `required_hash_fields` property read-only.

        This is a security mechanism such that the `required_hash_fields` property cannot be
        overridden by the user or malicious code.
        """
        if name == "body_hash":
            raise AttributeError(
                "body_hash property is read-only and cannot be overridden."
            )
        super().__setattr__(name, value)

    def get_total_size(self) -> int:
        """
        Get the total size of the current object.

        This method first calculates the size of the current object, then assigns it
        to the instance variable self.total_size and finally returns this value.

        Returns:
            int: The total size of the current object.
        """
        self.total_size = get_size(self)
        return self.total_size

    @property
    def is_success(self) -> bool:
        """
        Checks if the boson's status code indicates success.

        This method returns True if the status code of the boson is 200,
        which typically represents a successful HTTP request.

        Returns:
            bool: True if boson's status code is 200, False otherwise.
        """
        return self.boson.status_code == 200

    @property
    def is_failure(self) -> bool:
        """
        Checks if the boson's status code indicates failure.

        This method returns True if the status code of the boson is not 200,
        which would mean the HTTP request was not successful.

        Returns:
            bool: True if boson's status code is not 200, False otherwise.
        """
        return self.boson.status_code != 200

    @property
    def is_timeout(self) -> bool:
        """
        Checks if the boson's status code indicates a timeout.

        This method returns True if the status code of the boson is 408,
        which is the HTTP status code for a request timeout.

        Returns:
            bool: True if boson's status code is 408, False otherwise.
        """
        return self.boson.status_code == 408

    @property
    def is_blacklist(self) -> bool:
        """
        Checks if the boson's status code indicates a blacklisted request.

        This method returns True if the status code of the boson is 403,
        which is the HTTP status code for a forbidden request.

        Returns:
            bool: True if boson's status code is 403, False otherwise.
        """
        return self.boson.status_code == 403

    @property
    def is_failed_verification(self) -> bool:
        """
        Checks if the boson's status code indicates failed verification.

        This method returns True if the status code of the boson is 401,
        which is the HTTP status code for unauthorized access.

        Returns:
            bool: True if boson's status code is 401, False otherwise.
        """
        return self.boson.status_code == 401

    def to_headers(self) -> dict:
        """
        Converts the state of a Nucleon instance into a dictionary of HTTP headers. This method is essential for
        packaging Nucleon data for network transmission in the nimble framework, ensuring that each key aspect of
        the Nucleon is represented in a format suitable for HTTP communication.

        Process:
        1. Basic Information: It starts by including the 'name' and 'timeout' of the Nucleon, which are fundamental
            for identifying the query and managing its lifespan on the network.
        2. Complex Objects: The method serializes the 'fermion' and 'boson' objects, if present, into strings. This
            serialization is crucial for preserving the state and structure of these objects over the network.
        3. Encoding: Non-optional complex objects are serialized and encoded in base64, making them safe for HTTP
            transport.
        4. Size Metrics: The method calculates and adds the size of headers and the total object size, providing
            valuable information for network bandwidth management.

        Example Usage:
            nucleon = Nucleon(name="ExampleNucleon", timeout=30)
            headers = nucleon.to_headers()
            # headers now contains a dictionary representing the Nucleon instance

        Returns:
            dict: A dictionary containing key-value pairs representing the Nucleon's properties, suitable for HTTP communication.
        """
        # Initializing headers with 'name' and 'timeout'
        headers = {"name": self.name, "timeout": str(self.timeout)}

        # Adding headers for 'fermion' and 'boson' if they are not None
        headers.update(
            {
                f"nb_header_fermion_{k}": str(v)
                for k, v in self.fermion.dict().items()
                if v is not None
            }
        )
        headers.update(
            {
                f"nb_header_boson_{k}": str(v)
                for k, v in self.boson.dict().items()
                if v is not None
            }
        )

        # Getting the type hints for the properties of the instance
        property_type_hints = typing.get_type_hints(self)

        # Getting the fields of the instance
        instance_fields = self.dict()

        # Iterating over the fields of the instance
        for field, value in instance_fields.items():
            # If the object is not optional, serializing it, encoding it, and adding it to the headers
            required = schema([self.__class__])["definitions"][self.name].get(
                "required"
            )

            # Skipping the field if it's already in the headers or its value is None
            if field in headers or value is None:
                continue

            if required and field in required:
                try:
                    # create an empty (dummy) instance of type(value) to pass pydantic validation on the fermion side
                    serialized_value = json.dumps(value.__class__.__call__())
                    encoded_value = base64.b64encode(
                        serialized_value.encode()
                    ).decode("utf-8")
                    headers[f"nb_header_input_obj_{field}"] = encoded_value
                except TypeError as e:
                    raise ValueError(
                        f"Error serializing {field} with value {value}. Objects must be json serializable."
                    ) from e

        # Adding the size of the headers and the total size to the headers
        headers["header_size"] = str(sys.getsizeof(headers))
        headers["total_size"] = str(self.get_total_size())
        headers["computed_body_hash"] = self.body_hash

        return headers

    @property
    def body_hash(self) -> str:
        """
        Computes a SHA3-256 hash of the serialized body of the Nucleon instance. This hash is used to
        ensure the data integrity and security of the Nucleon instance when it's transmitted across the
        network. It's a crucial feature for verifying that the data received is the same as the data sent.

        Process:
        1. Iterates over each required field as specified in `required_fields_hash`.
        2. Concatenates the string representation of these fields.
        3. Applies SHA3-256 hashing to the concatenated string to produce a unique fingerprint of the data.

        Example:
            nucleon = Nucleon(name="ExampleRoute", timeout=10)
            hash_value = nucleon.body_hash
            # hash_value is the SHA3-256 hash of the serialized body of the Nucleon instance

        Returns:
            str: The SHA3-256 hash as a hexadecimal string, providing a fingerprint of the Nucleon instance's data for integrity checks.
        """
        # Hash the body for verification
        hashes = []

        # Getting the fields of the instance
        instance_fields = self.dict()

        for field, value in instance_fields.items():
            # If the field is required in the subclass schema, hash and add it.
            if field in self.required_hash_fields:
                hashes.append(nimble.utils.hash(str(value)))

        # Hash and return the hashes that have been concatenated
        return nimble.utils.hash("".join(hashes))

    @classmethod
    def parse_headers_to_inputs(cls, headers: dict) -> dict:
        """
        Interprets and transforms a given dictionary of headers into a structured dictionary, facilitating
        the reconstruction of Nucleon objects. This method is essential for parsing network-transmitted
        data back into a Nucleon instance, ensuring data consistency and integrity.

        Process:
        1. Separates headers into categories based on prefixes ('fermion', 'boson', etc.).
        2. Decodes and deserializes 'input_obj' headers into their original objects.
        3. Assigns simple fields directly from the headers to the input dictionary.

        Example:
            received_headers = {
                'nb_header_fermion_address': '127.0.0.1',
                'nb_header_boson_port': '8080',
                # Other headers...
            }
            inputs = Nucleon.parse_headers_to_inputs(received_headers)
            # inputs now contains a structured representation of Nucleon properties based on the headers

        Note: This is handled automatically when calling Nucleon.from_headers(headers) and does not
        need to be called directly.

        Args:
            headers (dict): The headers dictionary to parse.

        Returns:
            dict: A structured dictionary representing the inputs for constructing a Nucleon instance.
        """

        # Initialize the input dictionary with empty sub-dictionaries for 'fermion' and 'boson'
        inputs_dict = {"fermion": {}, "boson": {}}

        # Iterate over each item in the headers
        for key, value in headers.items():
            if "nb_header_fermion_" in key: # Handle 'fermion' headers
                try:
                    new_key = key.split("nb_header_fermion_")[1]
                    inputs_dict["fermion"][new_key] = value
                except Exception as e:
                    nimble.logging.error(
                        f"Error while parsing 'fermion' header {key}: {e}"
                    )
            elif "nb_header_boson_" in key: # Handle 'boson' headers
                try:
                    new_key = key.split("nb_header_boson_")[1]
                    inputs_dict["boson"][new_key] = value
                except Exception as e:
                    nimble.logging.error(
                        f"Error while parsing 'boson' header {key}: {e}"
                    )
            elif "nb_header_input_obj" in key: # Handle 'input_obj' headers
                try:
                    new_key = key.split("nb_header_input_obj_")[1]
                    # Skip if the key already exists in the dictionary
                    if new_key in inputs_dict:
                        continue
                    # Decode and load the serialized object
                    inputs_dict[new_key] = json.loads(
                        base64.b64decode(value.encode()).decode("utf-8")
                    )
                except json.JSONDecodeError as e:
                    nimble.logging.error(
                        f"Error while json decoding 'input_obj' header {key}: {e}"
                    )
                except Exception as e:
                    nimble.logging.error(
                        f"Error while parsing 'input_obj' header {key}: {e}"
                    )
            else: # Unexpected keys
                nimble.logging.info(
                    f"Unexpected key while parsing header with key {key}"
                )

        # Assign the remaining known headers directly
        inputs_dict["timeout"] = headers.get("timeout", None)
        inputs_dict["name"] = headers.get("name", None)
        inputs_dict["header_size"] = headers.get("header_size", None)
        inputs_dict["total_size"] = headers.get("total_size", None)
        inputs_dict["computed_body_hash"] = headers.get(
            "computed_body_hash", None
        )

        return inputs_dict

    @classmethod
    def from_headers(cls, headers: dict) -> "Nucleon":
        """
        Constructs a new Nucleon instance from a given headers dictionary, enabling the re-creation of the
        Nucleon's state as it was prior to network transmission. This method is a key part of the
        deserialization process in the nimble network, allowing nodes to accurately reconstruct Nucleon
        objects from received data.

        Example:
            received_headers = {
                'nb_header_fermion_address': '127.0.0.1',
                'nb_header_boson_port': '8080',
                # Other headers...
            }
            nucleon = Nucleon.from_headers(received_headers)
            # nucleon is a new Nucleon instance reconstructed from the received headers

        Args:
            headers (dict): The dictionary of headers containing serialized Nucleon information.

        Returns:
            Nucleon: A new instance of Nucleon, reconstructed from the parsed header information, replicating the original instance's state.
        """

        # Get the inputs dictionary from the headers
        input_dict = cls.parse_headers_to_inputs(headers)

        # Use the dictionary unpacking operator to pass the inputs to the class constructor
        nucleon = cls(**input_dict)

        return nucleon



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from munch import Munch, munchify

defaults: Munch = munchify(
    {
        "netuid": 1,
        "nbnetwork": {
            "network": "nimble",
            "chain_endpoint": None,
            "_mock": False,
        },
        "pow_register": {
            "num_processes": None,
            "update_interval": 50000,
            "output_in_place": True,
            "verbose": False,
            "cuda": {"dev_id": [0], "use_cuda": False, "tpb": 256},
        },
        "fermion": {
            "port": 8091,
            "ip": "[::]",
            "external_port": None,
            "external_ip": None,
            "max_workers": 10,
            "maximum_concurrent_rpcs": 400,
        },
        "priority": {"max_workers": 5, "maxsize": 10},
        "proton": {"port": 7091, "level": "INFO"},
        "wallet": {
            "name": "default",
            "hotkey": "default",
            "path": "~/.nimble/wallets/",
        },
        "dataset": {
            "batch_size": 10,
            "block_size": 20,
            "num_workers": 0,
            "dataset_names": "default",
            "data_dir": "~/.nimble/data/",
            "save_dataset": False,
            "max_datasets": 3,
            "num_batches": 100,
        },
        "logging": {
            "debug": False,
            "trace": False,
            "record_log": False,
            "logging_dir": "~/.nimble/miners",
        },
    }
)

from .stake import StakeCommand, StakeShow
from .unstake import UnStakeCommand
from .overview import OverviewCommand
from .list import ListCommand
from .wallets import (
    NewColdkeyCommand,
    NewHotkeyCommand,
    RegenColdkeyCommand,
    RegenColdkeypubCommand,
    RegenHotkeyCommand,
    UpdateWalletCommand,
    WalletCreateCommand,
    WalletBalanceCommand,
)
from .transfer import TransferCommand


# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

""" A exponential moving average that updates values based on time since last update.
"""

import time


class timed_rolling_avg:
    """A exponential moving average that updates values based on time since last update."""

    def __init__(self, initial_value, alpha):
        self.value = initial_value
        self.alpha = alpha
        self.last_update = time.time()

    def update(self, new_value):
        """Update self.value (the moving average) with the new_value"""
        now = time.time()
        time_delta = now - self.last_update
        self.last_update = now
        new_value = new_value / time_delta
        self.value = (1 - self.alpha) * self.value + self.alpha * new_value


class AmountPerSecondRollingAverage:
    """A exponential moving average that counts quantity per second."""

    def __init__(self, initial_value=0, alpha=0.1):
        self.value = initial_value
        self.alpha = alpha
        self.last_update = None

    def event(self, amount):
        """Update self.value (the moving average) with the new_value"""
        if self.last_update == None:
            self.last_update = time.time()
        else:
            now = time.time()
            time_delta = now - self.last_update
            self.last_update = now
            new_value = amount / time_delta
            self.value = (1 - self.alpha) * self.value + self.alpha * new_value

    def get(self) -> float:
        return float(self.value)


class EventsPerSecondRollingAverage:
    """A exponential moving average that counts the number of events per second."""

    def __init__(self, initial_value, alpha):
        self.value = initial_value
        self.alpha = alpha
        self.last_update = None

    def event(self):
        """Update self.value (the moving average) with the new_value"""
        if self.last_update == None:
            self.last_update = time.time()
        else:
            now = time.time()
            time_delta = now - self.last_update
            self.last_update = now
            new_value = 1 / time_delta
            self.value = (1 - self.alpha) * self.value + self.alpha * new_value

    def get(self) -> float:
        return float(self.value)



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import nimble

from rich.prompt import Confirm
from typing import List, Dict, Union
from ..utils.balance import Balance
from ..utils import is_valid_nimble_address_or_public_key


def transfer_extrinsic(
    nbnetwork: "nimble.nbnetwork",
    wallet: "nimble.wallet",
    dest: str,
    amount: Union[Balance, float],
    wait_for_inclusion: bool = True,
    wait_for_finalization: bool = False,
    keep_alive: bool = True,
    predict: bool = False,
) -> bool:
    r"""Transfers funds from this wallet to the destination public key address
    Args:
        wallet (nimble.wallet):
            nimble wallet object to make transfer from.
        dest (str, ss58_address or ed25519):
            Destination public key address of reciever.
        amount (Union[Balance, int]):
            Amount to stake as nimble balance, or float interpreted as NIM.
        wait_for_inclusion (bool):
            If set, waits for the extrinsic to enter a block before returning true,
            or returns false if the extrinsic fails to enter the block within the timeout.
        wait_for_finalization (bool):
            If set, waits for the extrinsic to be finalized on the chain before returning true,
            or returns false if the extrinsic fails to be finalized within the timeout.
        keep_alive (bool):
            If set, keeps the account alive by keeping the balance above the existential deposit.
        predict (bool):
            If true, the call waits for confirmation from the user before proceeding.
    Returns:
        success (bool):
            Flag is true if extrinsic was finalized or uncluded in the block.
            If we did not wait for finalization / inclusion, the response is true.
    """
    # Validate destination address.
    if not is_valid_nimble_address_or_public_key(dest):
        nimble.__console__.print(
            ":cross_mark: [red]Invalid destination address[/red]:[bold white]\n  {}[/bold white]".format(
                dest
            )
        )
        return False

    if isinstance(dest, bytes):
        # Convert bytes to hex string.
        dest = "0x" + dest.hex()

    # Unlock wallet coldkey.
    wallet.coldkey

    # Convert to nimble.Balance
    if not isinstance(amount, nimble.Balance):
        transfer_balance = nimble.Balance.from_nim(amount)
    else:
        transfer_balance = amount

    # Check balance.
    with nimble.__console__.status(":satellite: Checking Balance..."):
        account_balance = nbnetwork.get_balance(wallet.coldkey.ss58_address)
        # check existential deposit.
        existential_deposit = nbnetwork.get_existential_deposit()

    with nimble.__console__.status(":satellite: Transferring..."):
        fee = nbnetwork.get_transfer_fee(
            wallet=wallet, dest=dest, value=transfer_balance.vim
        )

    if not keep_alive:
        # Check if the transfer should keep_alive the account
        existential_deposit = nimble.Balance(0)

    # Check if we have enough balance.
    if account_balance < (transfer_balance + fee + existential_deposit):
        nimble.__console__.print(
            ":cross_mark: [red]Not enough balance[/red]:[bold white]\n  balance: {}\n  amount: {}\n  for fee: {}[/bold white]".format(
                account_balance, transfer_balance, fee
            )
        )
        return False

    # Ask before moving on.
    if predict:
        if not Confirm.ask(
            "Do you want to transfer:[bold white]\n  amount: {}\n  from: {}:{}\n  to: {}\n  for fee: {}[/bold white]".format(
                transfer_balance,
                wallet.name,
                wallet.coldkey.ss58_address,
                dest,
                fee,
            )
        ):
            return False

    with nimble.__console__.status(":satellite: Transferring..."):
        success, block_hash, err_msg = nbnetwork._do_transfer(
            wallet,
            dest,
            transfer_balance,
            wait_for_finalization=wait_for_finalization,
            wait_for_inclusion=wait_for_inclusion,
        )

        if success:
            nimble.__console__.print(
                ":white_heavy_check_mark: [green]Finalized[/green]"
            )
            nimble.__console__.print(
                "[green]Block Hash: {}[/green]".format(block_hash)
            )

            explorer_urls = nimble.utils.get_explorer_url_for_network(
                nbnetwork.network, block_hash, nimble.__network_explorer_map__
            )
            if explorer_urls != {}:
                nimble.__console__.print(
                    "[green]Nimble Explorer Link: {}[/green]".format(
                        explorer_urls.get("nimble-technology")
                    )
                )
                nimble.__console__.print(
                    "[green]Nimstats   Explorer Link: {}[/green]".format(
                        explorer_urls.get("nimstats")
                    )
                )
        else:
            nimble.__console__.print(
                ":cross_mark: [red]Failed[/red]: error:{}".format(err_msg)
            )

    if success:
        with nimble.__console__.status(":satellite: Checking Balance..."):
            new_balance = nbnetwork.get_balance(wallet.coldkey.ss58_address)
            nimble.__console__.print(
                "Balance:\n  [blue]{}[/blue] :arrow_right: [green]{}[/green]".format(
                    account_balance, new_balance
                )
            )
            return True

    return False



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import binascii
import hashlib
import math
import multiprocessing
import os
import random
import sys
import time
from dataclasses import dataclass
from datetime import timedelta
from queue import Empty, Full
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import backoff
import nimble
import torch
from Crypto.Hash import keccak
from rich import console as rich_console
from rich import status as rich_status

from .formatting import get_human_readable, millify
from ._register_cuda import solve_cuda


class CUDAException(Exception):
    """An exception raised when an error occurs in the CUDA environment."""

    pass


def _hex_bytes_to_u8_list(hex_bytes: bytes):
    hex_chunks = [
        int(hex_bytes[i : i + 2], 16) for i in range(0, len(hex_bytes), 2)
    ]
    return hex_chunks


def _create_seal_hash(block_and_hotkey_hash_bytes: bytes, nonce: int) -> bytes:
    nonce_bytes = binascii.hexlify(nonce.to_bytes(8, "little"))
    pre_seal = nonce_bytes + binascii.hexlify(block_and_hotkey_hash_bytes)[:64]
    seal_sh256 = hashlib.sha256(
        bytearray(_hex_bytes_to_u8_list(pre_seal))
    ).digest()
    kec = keccak.new(digest_bits=256)
    seal = kec.update(seal_sh256).digest()
    return seal


def _seal_meets_difficulty(seal: bytes, difficulty: int, limit: int):
    seal_number = int.from_bytes(seal, "big")
    product = seal_number * difficulty
    return product < limit


@dataclass
class POWSolution:
    """A solution to the registration PoW problem."""

    nonce: int
    block_number: int
    difficulty: int
    seal: bytes

    def is_stale(self, nbnetwork: "nimble.nbnetwork") -> bool:
        """Returns True if the POW is stale.
        This means the block the POW is solved for is within 3 blocks of the current block.
        """
        return self.block_number < nbnetwork.get_current_block() - 3


class _SolverBase(multiprocessing.Process):
    """
    A process that solves the registration PoW problem.

    Args:
        proc_num: int
            The number of the process being created.
        num_proc: int
            The total number of processes running.
        update_interval: int
            The number of nonces to try to solve before checking for a new block.
        finished_queue: multiprocessing.Queue
            The queue to put the process number when a process finishes each update_interval.
            Used for calculating the average time per update_interval across all processes.
        solution_queue: multiprocessing.Queue
            The queue to put the solution the process has found during the pow solve.
        newBlockEvent: multiprocessing.Event
            The event to set by the main process when a new block is finalized in the network.
            The solver process will check for the event after each update_interval.
            The solver process will get the new block hash and difficulty and start solving for a new nonce.
        stopEvent: multiprocessing.Event
            The event to set by the main process when all the solver processes should stop.
            The solver process will check for the event after each update_interval.
            The solver process will stop when the event is set.
            Used to stop the solver processes when a solution is found.
        curr_block: multiprocessing.Array
            The array containing this process's current block hash.
            The main process will set the array to the new block hash when a new block is finalized in the network.
            The solver process will get the new block hash from this array when newBlockEvent is set.
        curr_block_num: multiprocessing.Value
            The value containing this process's current block number.
            The main process will set the value to the new block number when a new block is finalized in the network.
            The solver process will get the new block number from this value when newBlockEvent is set.
        curr_diff: multiprocessing.Array
            The array containing this process's current difficulty.
            The main process will set the array to the new difficulty when a new block is finalized in the network.
            The solver process will get the new difficulty from this array when newBlockEvent is set.
        check_block: multiprocessing.Lock
            The lock to prevent this process from getting the new block data while the main process is updating the data.
        limit: int
            The limit of the pow solve for a valid solution.
    """

    proc_num: int
    num_proc: int
    update_interval: int
    finished_queue: multiprocessing.Queue
    solution_queue: multiprocessing.Queue
    newBlockEvent: multiprocessing.Event
    stopEvent: multiprocessing.Event
    hotkey_bytes: bytes
    curr_block: multiprocessing.Array
    curr_block_num: multiprocessing.Value
    curr_diff: multiprocessing.Array
    check_block: multiprocessing.Lock
    limit: int

    def __init__(
        self,
        proc_num,
        num_proc,
        update_interval,
        finished_queue,
        solution_queue,
        stopEvent,
        curr_block,
        curr_block_num,
        curr_diff,
        check_block,
        limit,
    ):
        multiprocessing.Process.__init__(self, daemon=True)
        self.proc_num = proc_num
        self.num_proc = num_proc
        self.update_interval = update_interval
        self.finished_queue = finished_queue
        self.solution_queue = solution_queue
        self.newBlockEvent = multiprocessing.Event()
        self.newBlockEvent.clear()
        self.curr_block = curr_block
        self.curr_block_num = curr_block_num
        self.curr_diff = curr_diff
        self.check_block = check_block
        self.stopEvent = stopEvent
        self.limit = limit

    def run(self):
        raise NotImplementedError("_SolverBase is an abstract class")

    @staticmethod
    def create_shared_memory() -> (
        Tuple[
            multiprocessing.Array, multiprocessing.Value, multiprocessing.Array
        ]
    ):
        """Creates shared memory for the solver processes to use."""
        curr_block = multiprocessing.Array("h", 32, lock=True)  # byte array
        curr_block_num = multiprocessing.Value("i", 0, lock=True)  # int
        curr_diff = multiprocessing.Array("Q", [0, 0], lock=True)  # [high, low]

        return curr_block, curr_block_num, curr_diff


class _Solver(_SolverBase):
    def run(self):
        block_number: int
        block_and_hotkey_hash_bytes: bytes
        block_difficulty: int
        nonce_limit = int(math.pow(2, 64)) - 1

        # Start at random nonce
        nonce_start = random.randint(0, nonce_limit)
        nonce_end = nonce_start + self.update_interval
        while not self.stopEvent.is_set():
            if self.newBlockEvent.is_set():
                with self.check_block:
                    block_number = self.curr_block_num.value
                    block_and_hotkey_hash_bytes = bytes(self.curr_block)
                    block_difficulty = _registration_diff_unpack(self.curr_diff)

                self.newBlockEvent.clear()

            # Do a block of nonces
            solution = _solve_for_nonce_block(
                nonce_start,
                nonce_end,
                block_and_hotkey_hash_bytes,
                block_difficulty,
                self.limit,
                block_number,
            )
            if solution is not None:
                self.solution_queue.put(solution)

            try:
                # Send time
                self.finished_queue.put_nowait(self.proc_num)
            except Full:
                pass

            nonce_start = random.randint(0, nonce_limit)
            nonce_start = nonce_start % nonce_limit
            nonce_end = nonce_start + self.update_interval


class _CUDASolver(_SolverBase):
    dev_id: int
    tpb: int

    def __init__(
        self,
        proc_num,
        num_proc,
        update_interval,
        finished_queue,
        solution_queue,
        stopEvent,
        curr_block,
        curr_block_num,
        curr_diff,
        check_block,
        limit,
        dev_id: int,
        tpb: int,
    ):
        super().__init__(
            proc_num,
            num_proc,
            update_interval,
            finished_queue,
            solution_queue,
            stopEvent,
            curr_block,
            curr_block_num,
            curr_diff,
            check_block,
            limit,
        )
        self.dev_id = dev_id
        self.tpb = tpb

    def run(self):
        block_number: int = 0  # dummy value
        block_and_hotkey_hash_bytes: bytes = b"0" * 32  # dummy value
        block_difficulty: int = int(math.pow(2, 64)) - 1  # dummy value
        nonce_limit = int(math.pow(2, 64)) - 1  # U64MAX

        # Start at random nonce
        nonce_start = random.randint(0, nonce_limit)
        while not self.stopEvent.is_set():
            if self.newBlockEvent.is_set():
                with self.check_block:
                    block_number = self.curr_block_num.value
                    block_and_hotkey_hash_bytes = bytes(self.curr_block)
                    block_difficulty = _registration_diff_unpack(self.curr_diff)

                self.newBlockEvent.clear()

            # Do a block of nonces
            solution = _solve_for_nonce_block_cuda(
                nonce_start,
                self.update_interval,
                block_and_hotkey_hash_bytes,
                block_difficulty,
                self.limit,
                block_number,
                self.dev_id,
                self.tpb,
            )
            if solution is not None:
                self.solution_queue.put(solution)

            try:
                # Signal that a nonce_block was finished using queue
                # send our proc_num
                self.finished_queue.put(self.proc_num)
            except Full:
                pass

            # increase nonce by number of nonces processed
            nonce_start += self.update_interval * self.tpb
            nonce_start = nonce_start % nonce_limit


def _solve_for_nonce_block_cuda(
    nonce_start: int,
    update_interval: int,
    block_and_hotkey_hash_bytes: bytes,
    difficulty: int,
    limit: int,
    block_number: int,
    dev_id: int,
    tpb: int,
) -> Optional[POWSolution]:
    """Tries to solve the POW on a CUDA device for a block of nonces (nonce_start, nonce_start + update_interval * tpb"""
    solution, seal = solve_cuda(
        nonce_start,
        update_interval,
        tpb,
        block_and_hotkey_hash_bytes,
        difficulty,
        limit,
        dev_id,
    )

    if solution != -1:
        # Check if solution is valid (i.e. not -1)
        return POWSolution(solution, block_number, difficulty, seal)

    return None


def _solve_for_nonce_block(
    nonce_start: int,
    nonce_end: int,
    block_and_hotkey_hash_bytes: bytes,
    difficulty: int,
    limit: int,
    block_number: int,
) -> Optional[POWSolution]:
    """Tries to solve the POW for a block of nonces (nonce_start, nonce_end)"""
    for nonce in range(nonce_start, nonce_end):
        # Create seal.
        seal = _create_seal_hash(block_and_hotkey_hash_bytes, nonce)

        # Check if seal meets difficulty
        if _seal_meets_difficulty(seal, difficulty, limit):
            # Found a solution, save it.
            return POWSolution(nonce, block_number, difficulty, seal)

    return None


def _registration_diff_unpack(packed_diff: multiprocessing.Array) -> int:
    """Unpacks the packed two 32-bit integers into one 64-bit integer. Little endian."""
    return int(packed_diff[0] << 32 | packed_diff[1])


def _registration_diff_pack(diff: int, packed_diff: multiprocessing.Array):
    """Packs the difficulty into two 32-bit integers. Little endian."""
    packed_diff[0] = diff >> 32
    packed_diff[1] = diff & 0xFFFFFFFF  # low 32 bits


def _hash_block_with_hotkey(block_bytes: bytes, hotkey_bytes: bytes) -> bytes:
    """Hashes the block with the hotkey using Keccak-256 to get 32 bytes"""
    kec = keccak.new(digest_bits=256)
    kec = kec.update(bytearray(block_bytes + hotkey_bytes))
    block_and_hotkey_hash_bytes = kec.digest()
    return block_and_hotkey_hash_bytes


def _update_curr_block(
    curr_diff: multiprocessing.Array,
    curr_block: multiprocessing.Array,
    curr_block_num: multiprocessing.Value,
    block_number: int,
    block_bytes: bytes,
    diff: int,
    hotkey_bytes: bytes,
    lock: multiprocessing.Lock,
):
    with lock:
        curr_block_num.value = block_number
        # Hash the block with the hotkey
        block_and_hotkey_hash_bytes = _hash_block_with_hotkey(
            block_bytes, hotkey_bytes
        )
        for i in range(32):
            curr_block[i] = block_and_hotkey_hash_bytes[i]
        _registration_diff_pack(diff, curr_diff)


def get_cpu_count() -> int:
    try:
        return len(os.sched_getaffinity(0))
    except AttributeError:
        # OSX does not have sched_getaffinity
        return os.cpu_count()


@dataclass
class RegistrationStatistics:
    """Statistics for a registration."""

    time_spent_total: float
    rounds_total: int
    time_average: float
    time_spent: float
    hash_rate_perpetual: float
    hash_rate: float
    difficulty: int
    block_number: int
    block_hash: bytes


class RegistrationStatisticsLogger:
    """Logs statistics for a registration."""

    console: rich_console.Console
    status: Optional[rich_status.Status]

    def __init__(
        self, console: rich_console.Console, output_in_place: bool = True
    ) -> None:
        self.console = console

        if output_in_place:
            self.status = self.console.status("Solving")
        else:
            self.status = None

    def start(self) -> None:
        if self.status is not None:
            self.status.start()

    def stop(self) -> None:
        if self.status is not None:
            self.status.stop()

    def get_status_message(
        cls, stats: RegistrationStatistics, verbose: bool = False
    ) -> str:
        message = (
            "Solving\n"
            + f"Time Spent (total): [bold white]{timedelta(seconds=stats.time_spent_total)}[/bold white]\n"
            + (
                f"Time Spent This Round: {timedelta(seconds=stats.time_spent)}\n"
                + f"Time Spent Average: {timedelta(seconds=stats.time_average)}\n"
                if verbose
                else ""
            )
            + f"Registration Difficulty: [bold white]{millify(stats.difficulty)}[/bold white]\n"
            + f"Iters (Inst/Perp): [bold white]{get_human_readable(stats.hash_rate, 'H')}/s / "
            + f"{get_human_readable(stats.hash_rate_perpetual, 'H')}/s[/bold white]\n"
            + f"Block Number: [bold white]{stats.block_number}[/bold white]\n"
            + f"Block Hash: [bold white]{stats.block_hash.encode('utf-8')}[/bold white]\n"
        )
        return message

    def update(
        self, stats: RegistrationStatistics, verbose: bool = False
    ) -> None:
        if self.status is not None:
            self.status.update(self.get_status_message(stats, verbose=verbose))
        else:
            self.console.log(self.get_status_message(stats, verbose=verbose))


def _solve_for_difficulty_fast(
    nbnetwork,
    wallet: "nimble.wallet",
    netuid: int,
    output_in_place: bool = True,
    num_processes: Optional[int] = None,
    update_interval: Optional[int] = None,
    n_samples: int = 10,
    alpha_: float = 0.80,
    log_verbose: bool = False,
) -> Optional[POWSolution]:
    """
    Solves the POW for registration using multiprocessing.
    Args:
        nbnetwork
            NbNetwork to connect to for block information and to submit.
        wallet:
            wallet to use for registration.
        netuid: int
            The netuid of the cosmos to register to.
        output_in_place: bool
            If true, prints the status in place. Otherwise, prints the status on a new line.
        num_processes: int
            Number of processes to use.
        update_interval: int
            Number of nonces to solve before updating block information.
        n_samples: int
            The number of samples of the hash_rate to keep for the EWMA
        alpha_: float
            The alpha for the EWMA for the hash_rate calculation
        log_verbose: bool
            If true, prints more verbose logging of the registration metrics.
    Note: The hash rate is calculated as an exponentially weighted moving average in order to make the measure more robust.
    Note:
    - We can also modify the update interval to do smaller blocks of work,
        while still updating the block information after a different number of nonces,
        to increase the transparency of the process while still keeping the speed.
    """
    if num_processes == None:
        # get the number of allowed processes for this process
        num_processes = min(1, get_cpu_count())

    if update_interval is None:
        update_interval = 50_000

    limit = int(math.pow(2, 256)) - 1

    curr_block, curr_block_num, curr_diff = _Solver.create_shared_memory()

    # Establish communication queues
    ## See the _Solver class for more information on the queues.
    stopEvent = multiprocessing.Event()
    stopEvent.clear()

    solution_queue = multiprocessing.Queue()
    finished_queues = [multiprocessing.Queue() for _ in range(num_processes)]
    check_block = multiprocessing.Lock()

    hotkey_bytes = (
        wallet.coldkeypub.public_key
        if netuid == -1
        else wallet.hotkey.public_key
    )
    # Start consumers
    solvers = [
        _Solver(
            i,
            num_processes,
            update_interval,
            finished_queues[i],
            solution_queue,
            stopEvent,
            curr_block,
            curr_block_num,
            curr_diff,
            check_block,
            limit,
        )
        for i in range(num_processes)
    ]

    # Get first block
    block_number, difficulty, block_hash = _get_block_with_retry(
        nbnetwork=nbnetwork, netuid=netuid
    )

    block_bytes = bytes.fromhex(block_hash[2:])
    old_block_number = block_number
    # Set to current block
    _update_curr_block(
        curr_diff,
        curr_block,
        curr_block_num,
        block_number,
        block_bytes,
        difficulty,
        hotkey_bytes,
        check_block,
    )

    # Set new block events for each solver to start at the initial block
    for worker in solvers:
        worker.newBlockEvent.set()

    for worker in solvers:
        worker.start()  # start the solver processes

    start_time = time.time()  # time that the registration started
    time_last = start_time  # time that the last work blocks completed

    curr_stats = RegistrationStatistics(
        time_spent_total=0.0,
        time_average=0.0,
        rounds_total=0,
        time_spent=0.0,
        hash_rate_perpetual=0.0,
        hash_rate=0.0,
        difficulty=difficulty,
        block_number=block_number,
        block_hash=block_hash,
    )

    start_time_perpetual = time.time()

    console = nimble.__console__
    logger = RegistrationStatisticsLogger(console, output_in_place)
    logger.start()

    solution = None

    hash_rates = [0] * n_samples  # The last n true hash_rates
    weights = [alpha_**i for i in range(n_samples)]  # weights decay by alpha

    while netuid == -1 or not nbnetwork.is_hotkey_registered(
        netuid=netuid, hotkey_ss58=wallet.hotkey.ss58_address
    ):
        # Wait until a solver finds a solution
        try:
            solution = solution_queue.get(block=True, timeout=0.25)
            if solution is not None:
                break
        except Empty:
            # No solution found, try again
            pass

        # check for new block
        old_block_number = _check_for_newest_block_and_update(
            nbnetwork=nbnetwork,
            netuid=netuid,
            hotkey_bytes=hotkey_bytes,
            old_block_number=old_block_number,
            curr_diff=curr_diff,
            curr_block=curr_block,
            curr_block_num=curr_block_num,
            curr_stats=curr_stats,
            update_curr_block=_update_curr_block,
            check_block=check_block,
            solvers=solvers,
        )

        num_time = 0
        for finished_queue in finished_queues:
            try:
                proc_num = finished_queue.get(timeout=0.1)
                num_time += 1

            except Empty:
                continue

        time_now = time.time()  # get current time
        time_since_last = (
            time_now - time_last
        )  # get time since last work block(s)
        if num_time > 0 and time_since_last > 0.0:
            # create EWMA of the hash_rate to make measure more robust

            hash_rate_ = (num_time * update_interval) / time_since_last
            hash_rates.append(hash_rate_)
            hash_rates.pop(0)  # remove the 0th data point
            curr_stats.hash_rate = sum(
                [hash_rates[i] * weights[i] for i in range(n_samples)]
            ) / (sum(weights))

            # update time last to now
            time_last = time_now

            curr_stats.time_average = (
                curr_stats.time_average * curr_stats.rounds_total
                + curr_stats.time_spent
            ) / (curr_stats.rounds_total + num_time)
            curr_stats.rounds_total += num_time

        # Update stats
        curr_stats.time_spent = time_since_last
        new_time_spent_total = time_now - start_time_perpetual
        curr_stats.hash_rate_perpetual = (
            curr_stats.rounds_total * update_interval
        ) / new_time_spent_total
        curr_stats.time_spent_total = new_time_spent_total

        # Update the logger
        logger.update(curr_stats, verbose=log_verbose)

    # exited while, solution contains the nonce or wallet is registered
    stopEvent.set()  # stop all other processes
    logger.stop()

    # terminate and wait for all solvers to exit
    _terminate_workers_and_wait_for_exit(solvers)

    return solution


@backoff.on_exception(backoff.constant, Exception, interval=1, max_tries=3)
def _get_block_with_retry(
    nbnetwork: "nimble.nbnetwork", netuid: int
) -> Tuple[int, int, bytes]:
    """
    Gets the current block number, difficulty, and block hash from the substrate node.

    Args:
        nbnetwork (:obj:`nimble.nbnetwork`, `required`):
            The nbnetwork object to use to get the block number, difficulty, and block hash.

        netuid (:obj:`int`, `required`):
            The netuid of the network to get the block number, difficulty, and block hash from.

    Returns:
        block_number (:obj:`int`):
            The current block number.

        difficulty (:obj:`int`):
            The current difficulty of the cosmos.

        block_hash (:obj:`bytes`):
            The current block hash.

    Raises:
        Exception: If the block hash is None.
        ValueError: If the difficulty is None.
    """
    block_number = nbnetwork.get_current_block()
    difficulty = (
        1_000_000 if netuid == -1 else nbnetwork.difficulty(netuid=netuid)
    )
    block_hash = nbnetwork.get_block_hash(block_number)
    if block_hash is None:
        raise Exception(
            "Network error. Could not connect to substrate to get block hash"
        )
    if difficulty is None:
        raise ValueError("Chain error. Difficulty is None")
    return block_number, difficulty, block_hash


class _UsingSpawnStartMethod:
    def __init__(self, force: bool = False):
        self._old_start_method = None
        self._force = force

    def __enter__(self):
        self._old_start_method = multiprocessing.get_start_method(
            allow_none=True
        )
        if self._old_start_method == None:
            self._old_start_method = "spawn"  # default to spawn

        multiprocessing.set_start_method("spawn", force=self._force)

    def __exit__(self, *args):
        # restore the old start method
        multiprocessing.set_start_method(self._old_start_method, force=True)


def _check_for_newest_block_and_update(
    nbnetwork: "nimble.nbnetwork",
    netuid: int,
    old_block_number: int,
    hotkey_bytes: bytes,
    curr_diff: multiprocessing.Array,
    curr_block: multiprocessing.Array,
    curr_block_num: multiprocessing.Value,
    update_curr_block: Callable,
    check_block: "multiprocessing.Lock",
    solvers: List[_Solver],
    curr_stats: RegistrationStatistics,
) -> int:
    """
    Checks for a new block and updates the current block information if a new block is found.

    Args:
        nbnetwork (:obj:`nimble.nbnetwork`, `required`):
            The nbnetwork object to use for getting the current block.
        netuid (:obj:`int`, `required`):
            The netuid to use for retrieving the difficulty.
        old_block_number (:obj:`int`, `required`):
            The old block number to check against.
        hotkey_bytes (:obj:`bytes`, `required`):
            The bytes of the hotkey's pubkey.
        curr_diff (:obj:`multiprocessing.Array`, `required`):
            The current difficulty as a multiprocessing array.
        curr_block (:obj:`multiprocessing.Array`, `required`):
            Where the current block is stored as a multiprocessing array.
        curr_block_num (:obj:`multiprocessing.Value`, `required`):
            Where the current block number is stored as a multiprocessing value.
        update_curr_block (:obj:`Callable`, `required`):
            A function that updates the current block.
        check_block (:obj:`multiprocessing.Lock`, `required`):
            A mp lock that is used to check for a new block.
        solvers (:obj:`List[_Solver]`, `required`):
            A list of solvers to update the current block for.
        curr_stats (:obj:`RegistrationStatistics`, `required`):
            The current registration statistics to update.

    Returns:
        (int) The current block number.
    """
    block_number = nbnetwork.get_current_block()
    if block_number != old_block_number:
        old_block_number = block_number
        # update block information
        block_number, difficulty, block_hash = _get_block_with_retry(
            nbnetwork=nbnetwork, netuid=netuid
        )
        block_bytes = bytes.fromhex(block_hash[2:])

        update_curr_block(
            curr_diff,
            curr_block,
            curr_block_num,
            block_number,
            block_bytes,
            difficulty,
            hotkey_bytes,
            check_block,
        )
        # Set new block events for each solver

        for worker in solvers:
            worker.newBlockEvent.set()

        # update stats
        curr_stats.block_number = block_number
        curr_stats.block_hash = block_hash
        curr_stats.difficulty = difficulty

    return old_block_number


def _solve_for_difficulty_fast_cuda(
    nbnetwork: "nimble.nbnetwork",
    wallet: "nimble.wallet",
    netuid: int,
    output_in_place: bool = True,
    update_interval: int = 50_000,
    tpb: int = 512,
    dev_id: Union[List[int], int] = 0,
    n_samples: int = 10,
    alpha_: float = 0.80,
    log_verbose: bool = False,
) -> Optional[POWSolution]:
    """
    Solves the registration fast using CUDA
    Args:
        nbnetwork: nimble.nbnetwork
            The nbnetwork node to grab blocks
        wallet: nimble.wallet
            The wallet to register
        netuid: int
            The netuid of the cosmos to register to.
        output_in_place: bool
            If true, prints the output in place, otherwise prints to new lines
        update_interval: int
            The number of nonces to try before checking for more blocks
        tpb: int
            The number of threads per block. CUDA param that should match the GPU capability
        dev_id: Union[List[int], int]
            The CUDA device IDs to execute the registration on, either a single device or a list of devices
        n_samples: int
            The number of samples of the hash_rate to keep for the EWMA
        alpha_: float
            The alpha for the EWMA for the hash_rate calculation
        log_verbose: bool
            If true, prints more verbose logging of the registration metrics.
    Note: The hash rate is calculated as an exponentially weighted moving average in order to make the measure more robust.
    """
    if isinstance(dev_id, int):
        dev_id = [dev_id]
    elif dev_id is None:
        dev_id = [0]

    if update_interval is None:
        update_interval = 50_000

    if not torch.cuda.is_available():
        raise Exception("CUDA not available")

    limit = int(math.pow(2, 256)) - 1

    # Set mp start to use spawn so CUDA doesn't complain
    with _UsingSpawnStartMethod(force=True):
        (
            curr_block,
            curr_block_num,
            curr_diff,
        ) = _CUDASolver.create_shared_memory()

        ## Create a worker per CUDA device
        num_processes = len(dev_id)

        # Establish communication queues
        stopEvent = multiprocessing.Event()
        stopEvent.clear()
        solution_queue = multiprocessing.Queue()
        finished_queues = [
            multiprocessing.Queue() for _ in range(num_processes)
        ]
        check_block = multiprocessing.Lock()

        hotkey_bytes = wallet.hotkey.public_key
        # Start workers
        solvers = [
            _CUDASolver(
                i,
                num_processes,
                update_interval,
                finished_queues[i],
                solution_queue,
                stopEvent,
                curr_block,
                curr_block_num,
                curr_diff,
                check_block,
                limit,
                dev_id[i],
                tpb,
            )
            for i in range(num_processes)
        ]

        # Get first block
        block_number, difficulty, block_hash = _get_block_with_retry(
            nbnetwork=nbnetwork, netuid=netuid
        )

        block_bytes = bytes.fromhex(block_hash[2:])
        old_block_number = block_number

        # Set to current block
        _update_curr_block(
            curr_diff,
            curr_block,
            curr_block_num,
            block_number,
            block_bytes,
            difficulty,
            hotkey_bytes,
            check_block,
        )

        # Set new block events for each solver to start at the initial block
        for worker in solvers:
            worker.newBlockEvent.set()

        for worker in solvers:
            worker.start()  # start the solver processes

        start_time = time.time()  # time that the registration started
        time_last = start_time  # time that the last work blocks completed

        curr_stats = RegistrationStatistics(
            time_spent_total=0.0,
            time_average=0.0,
            rounds_total=0,
            time_spent=0.0,
            hash_rate_perpetual=0.0,
            hash_rate=0.0,  # EWMA hash_rate (H/s)
            difficulty=difficulty,
            block_number=block_number,
            block_hash=block_hash,
        )

        start_time_perpetual = time.time()

        console = nimble.__console__
        logger = RegistrationStatisticsLogger(console, output_in_place)
        logger.start()

        hash_rates = [0] * n_samples  # The last n true hash_rates
        weights = [
            alpha_**i for i in range(n_samples)
        ]  # weights decay by alpha

        solution = None
        while netuid == -1 or not nbnetwork.is_hotkey_registered(
            netuid=netuid, hotkey_ss58=wallet.hotkey.ss58_address
        ):
            # Wait until a solver finds a solution
            try:
                solution = solution_queue.get(block=True, timeout=0.15)
                if solution is not None:
                    break
            except Empty:
                # No solution found, try again
                pass

            # check for new block
            old_block_number = _check_for_newest_block_and_update(
                nbnetwork=nbnetwork,
                netuid=netuid,
                hotkey_bytes=hotkey_bytes,
                curr_diff=curr_diff,
                curr_block=curr_block,
                curr_block_num=curr_block_num,
                old_block_number=old_block_number,
                curr_stats=curr_stats,
                update_curr_block=_update_curr_block,
                check_block=check_block,
                solvers=solvers,
            )

            num_time = 0
            # Get times for each solver
            for finished_queue in finished_queues:
                try:
                    proc_num = finished_queue.get(timeout=0.1)
                    num_time += 1

                except Empty:
                    continue

            time_now = time.time()  # get current time
            time_since_last = (
                time_now - time_last
            )  # get time since last work block(s)
            if num_time > 0 and time_since_last > 0.0:
                # create EWMA of the hash_rate to make measure more robust

                hash_rate_ = (
                    num_time * tpb * update_interval
                ) / time_since_last
                hash_rates.append(hash_rate_)
                hash_rates.pop(0)  # remove the 0th data point
                curr_stats.hash_rate = sum(
                    [hash_rates[i] * weights[i] for i in range(n_samples)]
                ) / (sum(weights))

                # update time last to now
                time_last = time_now

                curr_stats.time_average = (
                    curr_stats.time_average * curr_stats.rounds_total
                    + curr_stats.time_spent
                ) / (curr_stats.rounds_total + num_time)
                curr_stats.rounds_total += num_time

            # Update stats
            curr_stats.time_spent = time_since_last
            new_time_spent_total = time_now - start_time_perpetual
            curr_stats.hash_rate_perpetual = (
                curr_stats.rounds_total * (tpb * update_interval)
            ) / new_time_spent_total
            curr_stats.time_spent_total = new_time_spent_total

            # Update the logger
            logger.update(curr_stats, verbose=log_verbose)

        # exited while, found_solution contains the nonce or wallet is registered

        stopEvent.set()  # stop all other processes
        logger.stop()

        # terminate and wait for all solvers to exit
        _terminate_workers_and_wait_for_exit(solvers)

        return solution


def _terminate_workers_and_wait_for_exit(
    workers: List[multiprocessing.Process],
) -> None:
    for worker in workers:
        worker.terminate()
        worker.join()


def create_pow(
    nbnetwork,
    wallet,
    netuid: int,
    output_in_place: bool = True,
    cuda: bool = False,
    dev_id: Union[List[int], int] = 0,
    tpb: int = 256,
    num_processes: int = None,
    update_interval: int = None,
    log_verbose: bool = False,
) -> Optional[Dict[str, Any]]:
    """
    Creates a proof of work for the given nbnetwork and wallet.
    Args:
        nbnetwork (:obj:`nimble.nbnetwork.nbnetwork`, `required`):
            The nbnetwork to create a proof of work for.
        wallet (:obj:`nimble.wallet.wallet`, `required`):
            The wallet to create a proof of work for.
        netuid (:obj:`int`, `required`):
            The netuid for the cosmos to create a proof of work for.
        output_in_place (:obj:`bool`, `optional`, defaults to :obj:`True`):
            If true, prints the progress of the proof of work to the console
                in-place. Meaning the progress is printed on the same lines.
        cuda (:obj:`bool`, `optional`, defaults to :obj:`False`):
            If true, uses CUDA to solve the proof of work.
        dev_id (:obj:`Union[List[int], int]`, `optional`, defaults to :obj:`0`):
            The CUDA device id(s) to use. If cuda is true and dev_id is a list,
                then multiple CUDA devices will be used to solve the proof of work.
        tpb (:obj:`int`, `optional`, defaults to :obj:`256`):
            The number of threads per block to use when solving the proof of work.
            Should be a multiple of 32.
        num_processes (:obj:`int`, `optional`, defaults to :obj:`None`):
            The number of processes to use when solving the proof of work.
            If None, then the number of processes is equal to the number of
                CPU cores.
        update_interval (:obj:`int`, `optional`, defaults to :obj:`None`):
            The number of nonces to run before checking for a new block.
        log_verbose (:obj:`bool`, `optional`, defaults to :obj:`False`):
            If true, prints the progress of the proof of work more verbosely.
    Returns:
        :obj:`Optional[Dict[str, Any]]`: The proof of work solution or None if
            the wallet is already registered or there is a different error.

    Raises:
        :obj:`ValueError`: If the cosmos does not exist.
    """
    if netuid != -1:
        if not nbnetwork.cosmos_exists(netuid=netuid):
            raise ValueError(f"Cosmos {netuid} does not exist")

    if cuda:
        solution: Optional[POWSolution] = _solve_for_difficulty_fast_cuda(
            nbnetwork,
            wallet,
            netuid=netuid,
            output_in_place=output_in_place,
            dev_id=dev_id,
            tpb=tpb,
            update_interval=update_interval,
            log_verbose=log_verbose,
        )
    else:
        solution: Optional[POWSolution] = _solve_for_difficulty_fast(
            nbnetwork,
            wallet,
            netuid=netuid,
            output_in_place=output_in_place,
            num_processes=num_processes,
            update_interval=update_interval,
            log_verbose=log_verbose,
        )

    return solution



""" Standardize logging for nimble
"""
# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import copy
import nimble
import os
import sys
import torch

from loguru import logger

logger = logger.opt(colors=True)
# Remove default sink.
try:
    logger.remove(0)
except Exception:
    pass

import re


def _remove_loguru_ansi_directive(text: str) -> str:
    pattern = r"<.*?>"
    return re.sub(pattern, "", text)


class logging:
    """Standardize logging for nimble"""

    __has_been_inited__: bool = False
    __debug_on__: bool = False
    __trace_on__: bool = False
    __std_sink__: int = None
    __file_sink__: int = None

    def __new__(
        cls,
        config: "nimble.config" = None,
        debug: bool = None,
        trace: bool = None,
        record_log: bool = None,
        logging_dir: str = None,
    ):
        r"""Instantiate nimble logging system backend.
        Args:
            config (:obj:`nimble.config`, `optional`):
                nimble.logging.config()
            debug (:obj:`bool`, `optional`):
                Turn on debug.
            trace (:obj:`bool`, `optional`):
                Turn on trace.
            record_log (:obj:`bool`, `optional`):
                If true, logs are saved to loggind dir.
            logging_dir (:obj:`str`, `optional`):
                Directory where logs are sunk.
        """

        cls.__has_been_inited__ = True

        if config == None:
            config = logging.config()
        config = copy.deepcopy(config)
        config.logging.debug = debug if debug != None else config.logging.debug
        config.logging.trace = trace if trace != None else config.logging.trace
        config.logging.record_log = (
            record_log if record_log != None else config.logging.record_log
        )
        config.logging.logging_dir = (
            logging_dir if logging_dir != None else config.logging.logging_dir
        )

        # Remove default sink.
        try:
            logger.remove(0)
        except Exception:
            pass

        # Optionally Remove other sinks.
        if cls.__std_sink__ != None:
            logger.remove(cls.__std_sink__)
        if cls.__file_sink__ != None:
            logger.remove(cls.__file_sink__)

        # Add filtered sys.stdout.
        cls.__std_sink__ = logger.add(
            sys.stdout,
            level=0,
            filter=cls.log_filter,
            colorize=True,
            enqueue=True,
            backtrace=True,
            diagnose=True,
            format=cls.log_formatter,
        )

        cls.set_debug(config.logging.debug)
        cls.set_trace(config.logging.trace)

        # ---- Setup logging to root ----
        if config.logging.record_log:
            filepath = config.logging.logging_dir + "/logs.log"
            cls.__file_sink__ = logger.add(
                filepath,
                filter=cls.log_save_filter,
                enqueue=True,
                backtrace=True,
                diagnose=True,
                format=cls.log_save_formatter,
                rotation="25 MB",
                retention="10 days",
            )

    @classmethod
    def config(cls):
        """Get config from the argument parser
        Return: nimble.config object
        """
        parser = argparse.ArgumentParser()
        logging.add_args(parser)
        return nimble.config(parser, args=[])

    @classmethod
    def help(cls):
        """Print help to stdout"""
        parser = argparse.ArgumentParser()
        cls.add_args(parser)
        print(cls.__new__.__doc__)
        parser.print_help()

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser, prefix: str = None):
        """Accept specific arguments fro parser"""
        prefix_str = "" if prefix == None else prefix + "."
        try:
            default_logging_debug = os.getenv("NB_LOGGING_DEBUG") or False
            default_logging_trace = os.getenv("NB_LOGGING_TRACE") or False
            default_logging_record_log = (
                os.getenv("NB_LOGGING_RECORD_LOG") or False
            )
            default_logging_logging_dir = (
                os.getenv("NB_LOGGING_LOGGING_DIR") or "~/.nimble/miners"
            )
            parser.add_argument(
                "--" + prefix_str + "logging.debug",
                action="store_true",
                help="""Turn on nimble debugging information""",
                default=default_logging_debug,
            )
            parser.add_argument(
                "--" + prefix_str + "logging.trace",
                action="store_true",
                help="""Turn on nimble trace level information""",
                default=default_logging_trace,
            )
            parser.add_argument(
                "--" + prefix_str + "logging.record_log",
                action="store_true",
                help="""Turns on logging to file.""",
                default=default_logging_record_log,
            )
            parser.add_argument(
                "--" + prefix_str + "logging.logging_dir",
                type=str,
                help="Logging default root directory.",
                default=default_logging_logging_dir,
            )
        except argparse.ArgumentError:
            # re-parsing arguments.
            pass

    @classmethod
    def check_config(cls, config: "nimble.config"):
        """Check config"""
        assert config.logging

    @classmethod
    def set_debug(cls, debug_on: bool = True):
        """Set debug for the specific cls class"""
        if not cls.__has_been_inited__:
            cls()
        cls.__debug_on__ = debug_on

    @classmethod
    def set_trace(cls, trace_on: bool = True):
        """Set trace back for the specific cls class"""
        if not cls.__has_been_inited__:
            cls()
        cls.__trace_on__ = trace_on

    @classmethod
    def get_level(cls) -> int:
        return 5 if cls.__trace_on__ else 10 if cls.__debug_on__ else 20

    @classmethod
    def log_filter(cls, record):
        """Filter out debug log if debug is not on"""
        if cls.get_level() <= record["level"].no:
            return True
        else:
            return False

    @classmethod
    def log_save_filter(cls, record):
        if cls.get_level() < record["level"].no:
            return True
        else:
            return False

    @classmethod
    def log_formatter(cls, record):
        """Log with different format according to record['extra']"""
        return "<blue>{time:YYYY-MM-DD HH:mm:ss.SSS}</blue> | <level>{level: ^16}</level> | {message}\n"

    @classmethod
    def log_save_formatter(cls, record):
        if cls.__trace_on__:
            return "{time:YYYY-MM-DD HH:mm:ss.SSS} | <level>{level: ^16}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | {message}\n"
        else:
            return "{time:YYYY-MM-DD HH:mm:ss.SSS} | <level>{level: ^16}</level> | {message}\n"

    @classmethod
    def _format(cls, prefix: object, sufix: object = None):
        """Format logging message"""
        if isinstance(prefix, torch.Tensor):
            prefix = prefix.detach()
        if sufix != None:
            if isinstance(sufix, torch.Tensor):
                sufix = "shape: {}".format(
                    str(sufix.shape)
                ) + " data: {}".format(str(sufix.detach()))
            else:
                sufix = "{}".format(str(sufix))
        else:
            sufix = ""
        log_msg = str(prefix).ljust(30) + str(sufix)
        return _remove_loguru_ansi_directive(log_msg)

    @classmethod
    def success(cls, prefix: object, sufix: object = None):
        """Success logging"""
        if not cls.__has_been_inited__:
            cls()
        logger.success(cls._format(prefix, sufix))

    @classmethod
    def warning(cls, prefix: object, sufix: object = None):
        """Warning logging"""
        if not cls.__has_been_inited__:
            cls()
        logger.warning(cls._format(prefix, sufix))

    @classmethod
    def error(cls, prefix: object, sufix: object = None):
        """Error logging"""
        if not cls.__has_been_inited__:
            cls()
        logger.error(cls._format(prefix, sufix))

    @classmethod
    def info(cls, prefix: object, sufix: object = None):
        """Info logging"""
        if not cls.__has_been_inited__:
            cls()
        logger.info(cls._format(prefix, sufix))

    @classmethod
    def debug(cls, prefix: object, sufix: object = None):
        """Info logging"""
        if not cls.__has_been_inited__:
            cls()
        logger.debug(cls._format(prefix, sufix))

    @classmethod
    def trace(cls, prefix: object, sufix: object = None):
        """Info logging"""
        if not cls.__has_been_inited__:
            cls()
        logger.trace(cls._format(prefix, sufix))

    @classmethod
    def exception(cls, prefix: object, sufix: object = None):
        """Exception logging with traceback"""
        if not cls.__has_been_inited__:
            cls()
        logger.exception(cls._format(prefix, sufix))



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import binascii
import hashlib
import math
from typing import Tuple

import numpy as np
from Crypto.Hash import keccak

from contextlib import redirect_stdout
import io


def solve_cuda(
    nonce_start: np.int64,
    update_interval: np.int64,
    tpb: int,
    block_and_hotkey_hash_bytes: bytes,
    difficulty: int,
    limit: int,
    dev_id: int = 0,
) -> Tuple[np.int64, bytes]:
    """
    Solves the PoW problem using CUDA.
    Args:
        nonce_start: int64
            Starting nonce.
        update_interval: int64
            Number of nonces to solve before updating block information.
        tpb: int
            Threads per block.
        block_and_hotkey_hash_bytes: bytes
            Keccak(Bytes of the block hash + bytes of the hotkey) 64 bytes.
        difficulty: int256
            Difficulty of the PoW problem.
        limit: int256
            Upper limit of the nonce.
        dev_id: int (default=0)
            The CUDA device ID
    Returns:
        Tuple[int64, bytes]
            Tuple of the nonce and the seal corresponding to the solution.
            Returns -1 for nonce if no solution is found.
    """

    try:
        import cubit
    except ImportError:
        raise ImportError("Please install cubit")

    upper = int(limit // difficulty)

    upper_bytes = upper.to_bytes(32, byteorder="little", signed=False)

    def _hex_bytes_to_u8_list(hex_bytes: bytes):
        hex_chunks = [
            int(hex_bytes[i : i + 2], 16) for i in range(0, len(hex_bytes), 2)
        ]
        return hex_chunks

    def _create_seal_hash(
        block_and_hotkey_hash_hex: bytes, nonce: int
    ) -> bytes:
        nonce_bytes = binascii.hexlify(nonce.to_bytes(8, "little"))
        pre_seal = nonce_bytes + block_and_hotkey_hash_hex
        seal_sh256 = hashlib.sha256(
            bytearray(_hex_bytes_to_u8_list(pre_seal))
        ).digest()
        kec = keccak.new(digest_bits=256)
        seal = kec.update(seal_sh256).digest()
        return seal

    def _seal_meets_difficulty(seal: bytes, difficulty: int):
        seal_number = int.from_bytes(seal, "big")
        product = seal_number * difficulty
        limit = int(math.pow(2, 256)) - 1

        return product < limit

    # Call cython function
    # int blockSize, uint64 nonce_start, uint64 update_interval, const unsigned char[:] limit,
    # const unsigned char[:] block_bytes, int dev_id
    block_and_hotkey_hash_hex = binascii.hexlify(block_and_hotkey_hash_bytes)[
        :64
    ]

    solution = cubit.solve_cuda(
        tpb,
        nonce_start,
        update_interval,
        upper_bytes,
        block_and_hotkey_hash_hex,
        dev_id,
    )  # 0 is first GPU
    seal = None
    if solution != -1:
        seal = _create_seal_hash(block_and_hotkey_hash_hex, solution)
        if _seal_meets_difficulty(seal, difficulty):
            return solution, seal
        else:
            return -1, b"\x00" * 32

    return solution, seal


def reset_cuda():
    """
    Resets the CUDA environment.
    """
    try:
        import cubit
    except ImportError:
        raise ImportError("Please install cubit")

    cubit.reset_cuda()


def log_cuda_errors() -> str:
    """
    Logs any CUDA errors.
    """
    try:
        import cubit
    except ImportError:
        raise ImportError("Please install cubit")

    f = io.StringIO()
    with redirect_stdout(f):
        cubit.log_cuda_errors()

    s = f.getvalue()

    return s



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
import argparse
import nimble
from rich.prompt import Prompt
from . import defaults

console = nimble.__console__


class TransferCommand:
    """
    Executes the 'transfer' command to transfer NIM tokens from one account to another on the nimble network.
    This command is used for transactions between different accounts, enabling users to send tokens to other participants on the network.

    Usage:
    The command requires specifying the destination address (public key) and the amount of NIM to be transferred.
    It checks for sufficient balance and prompts for confirmation before proceeding with the transaction.

    Optional arguments:
    - --dest (str): The destination address for the transfer. This can be in the form of an SS58 or ed2519 public key.
    - --amount (float): The amount of NIM tokens to transfer.

    The command displays the user's current balance before inference for the amount to transfer, ensuring transparency and accuracy in the transaction.

    Example usage:
    >>> nbcli wallet transfer --dest 5Dp8... --amount 100

    Note:
    This command is crucial for executing token transfers within the nimble network. Users should verify the destination address and amount
    before confirming the transaction to avoid errors or loss of funds.
    """

    @staticmethod
    def run(cli):
        r"""Transfer token of amount to destination."""
        wallet = nimble.wallet(config=cli.config)
        nbnetwork = nimble.nbnetwork(config=cli.config, log_verbose=False)
        nbnetwork.transfer(
            wallet=wallet,
            dest=cli.config.dest,
            amount=cli.config.amount,
            wait_for_inclusion=True,
            predict=not cli.config.no_predict,
        )

    @staticmethod
    def check_config(config: "nimble.config"):
        if not config.is_set("wallet.name") and not config.no_predict:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)

        # Get destination.
        if not config.dest and not config.no_predict:
            dest = Prompt.ask("Enter destination public key: (ss58 or ed2519)")
            if not nimble.utils.is_valid_nimble_address_or_public_key(dest):
                sys.exit()
            else:
                config.dest = str(dest)

        # Get current balance and print to user.
        if not config.no_predict:
            wallet = nimble.wallet(config=config)
            nbnetwork = nimble.nbnetwork(config=config, log_verbose=False)
            with nimble.__console__.status(":satellite: Checking Balance..."):
                account_balance = nbnetwork.get_balance(
                    wallet.coldkeypub.ss58_address
                )
                nimble.__console__.print(
                    "Balance: [green]{}[/green]".format(account_balance)
                )

        # Get amount.
        if not config.get("amount"):
            if not config.no_predict:
                amount = Prompt.ask("Enter NIM amount to transfer")
                try:
                    config.amount = float(amount)
                except ValueError:
                    console.print(
                        ":cross_mark:[red] Invalid NIM amount[/red] [bold white]{}[/bold white]".format(
                            amount
                        )
                    )
                    sys.exit()
            else:
                console.print(
                    ":cross_mark:[red] Invalid NIM amount[/red] [bold white]{}[/bold white]".format(
                        amount
                    )
                )
                sys.exit(1)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        transfer_parser = parser.add_parser(
            "transfer", help="""Transfer NIM between accounts."""
        )
        transfer_parser.add_argument(
            "--dest", dest="dest", type=str, required=False
        )
        transfer_parser.add_argument(
            "--amount", dest="amount", type=float, required=False
        )

        nimble.wallet.add_args(transfer_parser)
        nimble.nbnetwork.add_args(transfer_parser)



# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
import argparse
import nimble
from rich.prompt import Prompt, Confirm
from .utils import check_netuid_set

from . import defaults

console = nimble.__console__


class RegisterCommand:
    """
    Executes the 'register' command to register a neuron on the nimble network by recycling some NIM (the network's native token).
    This command is used to add a new neuron to a specified subnet within the network, contributing to the decentralization and robustness of nimble.

    Usage:
    Before registering, the command checks if the specified subnet exists and whether the user's balance is sufficient to cover the registration cost.
    The registration cost is determined by the current recycle amount for the specified subnet. If the balance is insufficient or the subnet does not exist,
    the command will exit with an appropriate error message.

    If the preconditions are met, and the user confirms the transaction (if 'no_prompt' is not set), the command proceeds to register the neuron by burning the required amount of NIM.

    The command structure includes:
    - Verification of subnet existence.
    - Checking the user's balance against the current recycle amount for the subnet.
    - User confirmation prompt for proceeding with registration.
    - Execution of the registration process.

    Columns Displayed in the Confirmation Prompt:
    - Balance: The current balance of the user's wallet in NIM.
    - Cost to Register: The required amount of NIM needed to register on the specified subnet.

    Example usage:
    >>> nbcli subnets register --netuid 1

    Note:
    This command is critical for users who wish to contribute a new neuron to the network. It requires careful consideration of the subnet selection and
    an understanding of the registration costs. Users should ensure their wallet is sufficiently funded before attempting to register a neuron.
    """

    @staticmethod
    def run(cli):
        r"""Register neuron by recycling some NIM."""
        wallet = nimble.wallet(config=cli.config)
        nbnetwork = nimble.nbnetwork(config=cli.config, log_verbose=False)

        # Verify subnet exists
        if not nbnetwork.cosmos_exists(netuid=cli.config.netuid):
            nimble.__console__.print(
                f"[red]Subnet {cli.config.netuid} does not exist[/red]"
            )
            sys.exit(1)

        # Check current recycle amount
        current_recycle = nbnetwork.burn(netuid=cli.config.netuid)
        balance = nbnetwork.get_balance(address=wallet.coldkeypub.ss58_address)

        # Check balance is sufficient
        if balance < current_recycle:
            nimble.__console__.print(
                f"[red]Insufficient balance {balance} to register neuron. Current recycle is {current_recycle} NIM[/red]"
            )
            sys.exit(1)

        if not cli.config.no_prompt:
            if (
                Confirm.ask(
                    f"Your balance is: [bold green]{balance}[/bold green]\nThe cost to register by recycle is [bold red]{current_recycle}[/bold red]\nDo you want to continue?",
                    default=False,
                )
                == False
            ):
                sys.exit(1)

        nbnetwork.burned_register(
            wallet=wallet,
            netuid=cli.config.netuid,
            prompt=not cli.config.no_prompt,
        )

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        register_parser = parser.add_parser(
            "register", help="""Register a wallet to a network."""
        )
        register_parser.add_argument(
            "--netuid",
            type=int,
            help="netuid for subnet to serve this neuron on",
            default=argparse.SUPPRESS,
        )

        nimble.wallet.add_args(register_parser)
        nimble.nbnetwork.add_args(register_parser)

    @staticmethod
    def check_config(config: "nimble.config"):
        if (
            not config.is_set("nbnetwork.network")
            and not config.is_set("nbnetwork.chain_endpoint")
            and not config.no_prompt
        ):
            config.nbnetwork.network = Prompt.ask(
                "Enter nbnetwork network",
                choices=nimble.__networks__,
                default=defaults.nbnetwork.network,
            )
            _, endpoint = nimble.nbnetwork.determine_chain_endpoint_and_network(
                config.nbnetwork.network
            )
            config.nbnetwork.chain_endpoint = endpoint

        check_netuid_set(
            config, nbnetwork=nimble.nbnetwork(config=config, log_verbose=False)
        )

        if not config.is_set("wallet.name") and not config.no_prompt:
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)

        if not config.is_set("wallet.hotkey") and not config.no_prompt:
            hotkey = Prompt.ask(
                "Enter hotkey name", default=defaults.wallet.hotkey
            )
            config.wallet.hotkey = str(hotkey)


# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import nimble
from tqdm import tqdm
from concurrent.futures import ProcessPoolExecutor
from collections import defaultdict
from fuzzywuzzy import fuzz
from rich.align import Align
from rich.table import Table
from rich.prompt import Prompt
from typing import List, Optional, Dict, Tuple
from .utils import (
    get_hotkey_wallets_for_wallet,
    get_coldkey_wallets_for_path,
    get_all_wallets_for_path,
)
from . import defaults

console = nimble.__console__


class OverviewCommand:
    """
    Executes the 'overview' command to present a detailed overview of the user's registered accounts on the nimble network.
    This command compiles and displays comprehensive information about each particle associated with the user's wallets,
    including both hotkeys and coldkeys. It is especially useful for users managing multiple accounts or seeking a summary
    of their network activities and stake distributions.

    Usage:
    The command offers various options to customize the output. Users can filter the displayed data by specific netuids,
    sort by different criteria, and choose to include all wallets in the user's configuration directory. The output is
    presented in a tabular format with the following columns:
    - COLDKEY: The SS58 address of the coldkey.
    - HOTKEY: The SS58 address of the hotkey.
    - UID: Unique identifier of the particle.
    - ACTIVE: Indicates if the particle is active.
    - STAKE(τ): Amount of stake in the particle, in Nim.
    - RANK: The rank of the particle within the network.
    - TRUST: Trust score of the particle.
    - CONSENSUS: Consensus score of the particle.
    - INCENTIVE: Incentive score of the particle.
    - DIVIDENDS: Dividends earned by the particle.
    - EMISSION(p): Emission received by the particle, in Rho.
    - VTRUST: Validator trust score of the particle.
    - VPERMIT: Indicates if the particle has a validator permit.
    - UPDATED: Time since last update.
    - FERMION: IP address and port of the particle.
    - HOTKEY_SS58: Human-readable representation of the hotkey.

    Example usage:
    >>> nbcli wallet overview
    >>> nbcli wallet overview --all --sort_by stake --sort_order descending

    Note:
    This command is read-only and does not modify the network state or account configurations. It provides a quick and
    comprehensive view of the user's network presence, making it ideal for monitoring account status, stake distribution,
    and overall contribution to the nimble network.
    """

    @staticmethod
    def run(cli: "nimble.cli"):
        r"""Prints an overview for the wallet's colkey."""
        console = nimble.__console__
        wallet = nimble.wallet(config=cli.config)
        nbnetwork: "nimble.nbnetwork" = nimble.nbnetwork(
            config=cli.config, log_verbose=False
        )

        all_hotkeys = []
        total_balance = nimble.Balance(0)

        # We are printing for every coldkey.
        if cli.config.get("all", d=None):
            cold_wallets = get_coldkey_wallets_for_path(cli.config.wallet.path)
            for cold_wallet in tqdm(cold_wallets, desc="Pulling balances"):
                if (
                    cold_wallet.coldkeypub_file.exists_on_device()
                    and not cold_wallet.coldkeypub_file.is_encrypted()
                ):
                    total_balance = total_balance + nbnetwork.get_balance(
                        cold_wallet.coldkeypub.ss58_address
                    )
            all_hotkeys = get_all_wallets_for_path(cli.config.wallet.path)
        else:
            # We are only printing keys for a single coldkey
            coldkey_wallet = nimble.wallet(config=cli.config)
            if (
                coldkey_wallet.coldkeypub_file.exists_on_device()
                and not coldkey_wallet.coldkeypub_file.is_encrypted()
            ):
                total_balance = nbnetwork.get_balance(
                    coldkey_wallet.coldkeypub.ss58_address
                )
            if not coldkey_wallet.coldkeypub_file.exists_on_device():
                console.print("[bold red]No wallets found.")
                return
            all_hotkeys = get_hotkey_wallets_for_wallet(coldkey_wallet)

        # We are printing for a select number of hotkeys from all_hotkeys.

        if cli.config.get("hotkeys", []):
            if not cli.config.get("all_hotkeys", False):
                # We are only showing hotkeys that are specified.
                all_hotkeys = [
                    hotkey
                    for hotkey in all_hotkeys
                    if hotkey.hotkey_str in cli.config.hotkeys
                ]
            else:
                # We are excluding the specified hotkeys from all_hotkeys.
                all_hotkeys = [
                    hotkey
                    for hotkey in all_hotkeys
                    if hotkey.hotkey_str not in cli.config.hotkeys
                ]

        # Check we have keys to display.
        if len(all_hotkeys) == 0:
            console.print("[red]No wallets found.[/red]")
            return

        # Pull particle info for all keys.
        particles: Dict[str, List[nimble.ParticleInfoLite]] = {}
        block = nbnetwork.block

        netuids = nbnetwork.get_all_cosmos_netuids()
        if cli.config.netuid != []:
            netuids = [
                netuid for netuid in netuids if netuid in cli.config.netuid
            ]
        for netuid in netuids:
            particles[str(netuid)] = []

        all_wallet_names = set([wallet.name for wallet in all_hotkeys])
        all_coldkey_wallets = [
            nimble.wallet(name=wallet_name) for wallet_name in all_wallet_names
        ]

        hotkey_coldkey_to_hotkey_wallet = {}
        for hotkey_wallet in all_hotkeys:
            if (
                hotkey_wallet.hotkey.ss58_address
                not in hotkey_coldkey_to_hotkey_wallet
            ):
                hotkey_coldkey_to_hotkey_wallet[
                    hotkey_wallet.hotkey.ss58_address
                ] = {}

            hotkey_coldkey_to_hotkey_wallet[hotkey_wallet.hotkey.ss58_address][
                hotkey_wallet.coldkeypub.ss58_address
            ] = hotkey_wallet

        all_hotkey_addresses = list(hotkey_coldkey_to_hotkey_wallet.keys())
        with console.status(
            ":satellite: Syncing with chain: [white]{}[/white] ...".format(
                cli.config.nbnetwork.get(
                    "network", nimble.defaults.nbnetwork.network
                )
            )
        ):
            # Create a copy of the config without the parser and formatter_class.
            ## This is needed to pass to the ProcessPoolExecutor, which cannot pickle the parser.
            copy_config = cli.config.copy()
            copy_config["__parser"] = None
            copy_config["formatter_class"] = None

            # Pull particle info for all keys.
            ## Max len(netuids) or 5 threads.
            with ProcessPoolExecutor(
                max_workers=max(len(netuids), 5)
            ) as executor:
                results = executor.map(
                    OverviewCommand._get_particles_for_netuid,
                    [
                        (copy_config, netuid, all_hotkey_addresses)
                        for netuid in netuids
                    ],
                )
                executor.shutdown(wait=True)  # wait for all complete

                for result in results:
                    netuid, particles_result, err_msg = result
                    if err_msg is not None:
                        console.print(err_msg)

                    if len(particles_result) == 0:
                        # Remove netuid from overview if no particles are found.
                        netuids.remove(netuid)
                        del particles[str(netuid)]
                    else:
                        # Add particles to overview.
                        particles[str(netuid)] = particles_result

            total_coldkey_stake_from_megastring = defaultdict(
                lambda: nimble.Balance(0.0)
            )
            checked_hotkeys = set()
            for particle_list in particles.values():
                for particle in particle_list:
                    if particle.hotkey in checked_hotkeys:
                        continue
                    total_coldkey_stake_from_megastring[
                        particle.coldkey
                    ] += particle.stake_dict[particle.coldkey]
                    checked_hotkeys.add(particle.hotkey)

            alerts_table = Table(show_header=True, header_style="bold magenta")
            alerts_table.add_column("🥩 alert!")

            coldkeys_to_check = []
            for coldkey_wallet in all_coldkey_wallets:
                # Check if we have any stake with hotkeys that are not registered.
                total_coldkey_stake_from_chain = (
                    nbnetwork.get_total_stake_for_coldkey(
                        ss58_address=coldkey_wallet.coldkeypub.ss58_address
                    )
                )
                difference = (
                    total_coldkey_stake_from_chain
                    - total_coldkey_stake_from_megastring[
                        coldkey_wallet.coldkeypub.ss58_address
                    ]
                )
                if difference == 0:
                    continue  # We have all our stake registered.

                coldkeys_to_check.append(coldkey_wallet)
                alerts_table.add_row(
                    "Found {} stake with coldkey {} that is not registered.".format(
                        difference, coldkey_wallet.coldkeypub.ss58_address
                    )
                )

            if len(coldkeys_to_check) > 0:
                # We have some stake that is not with a registered hotkey.
                if "-1" not in particles:
                    particles["-1"] = []

            # Use process pool to check each coldkey wallet for de-registered stake.
            with ProcessPoolExecutor(
                max_workers=max(len(coldkeys_to_check), 5)
            ) as executor:
                results = executor.map(
                    OverviewCommand._get_de_registered_stake_for_coldkey_wallet,
                    [
                        (cli.config, all_hotkey_addresses, coldkey_wallet)
                        for coldkey_wallet in coldkeys_to_check
                    ],
                )
                executor.shutdown(wait=True)  # wait for all complete

            for result in results:
                coldkey_wallet, de_registered_stake, err_msg = result
                if err_msg is not None:
                    console.print(err_msg)

                if len(de_registered_stake) == 0:
                    continue  # We have no de-registered stake with this coldkey.

                de_registered_particles = []
                for hotkey_addr, our_stake in de_registered_stake:
                    # Make a particle info lite for this hotkey and coldkey.
                    de_registered_particle = nimble.ParticleInfoLite._null_particle()
                    de_registered_particle.hotkey = hotkey_addr
                    de_registered_particle.coldkey = (
                        coldkey_wallet.coldkeypub.ss58_address
                    )
                    de_registered_particle.total_stake = nimble.Balance(our_stake)

                    de_registered_particles.append(de_registered_particle)

                    # Add this hotkey to the wallets dict
                    wallet_ = nimble.Wallet(
                        name=wallet,
                    )
                    wallet_.hotkey = hotkey_addr
                    wallet.hotkey_str = hotkey_addr[
                        :5
                    ]  # Max length of 5 characters
                    hotkey_coldkey_to_hotkey_wallet[hotkey_addr][
                        coldkey_wallet.coldkeypub.ss58_address
                    ] = wallet_

                # Add particles to overview.
                particles["-1"].extend(de_registered_particles)

        # Setup outer table.
        grid = Table.grid(pad_edge=False)

        # If there are any alerts, add them to the grid
        if len(alerts_table.rows) > 0:
            grid.add_row(alerts_table)

        title: str = ""
        if not cli.config.get("all", d=None):
            title = "[bold white italic]Wallet - {}:{}".format(
                cli.config.wallet.name, wallet.coldkeypub.ss58_address
            )
        else:
            title = "[bold whit italic]All Wallets:"

        # Add title
        grid.add_row(Align(title, vertical="middle", align="center"))

        # Generate rows per netuid
        hotkeys_seen = set()
        total_particles = 0
        total_stake = 0.0
        for netuid in netuids:
            cosmos_tempo = nbnetwork.tempo(netuid=netuid)
            last_cosmos = netuid == netuids[-1]
            TABLE_DATA = []
            total_rank = 0.0
            total_trust = 0.0
            total_consensus = 0.0
            total_validator_trust = 0.0
            total_incentive = 0.0
            total_dividends = 0.0
            total_emission = 0

            for nn in particles[str(netuid)]:
                hotwallet = hotkey_coldkey_to_hotkey_wallet.get(
                    nn.hotkey, {}
                ).get(nn.coldkey, None)
                if not hotwallet:
                    # Indicates a mismatch between what the chain says the coldkey
                    # is for this hotkey and the local wallet coldkey-hotkey pair
                    hotwallet = argparse.Namespace()
                    hotwallet.name = nn.coldkey[:7]
                    hotwallet.hotkey_str = nn.hotkey[:7]
                nn: nimble.ParticleInfoLite
                uid = nn.uid
                active = nn.active
                stake = nn.total_stake.nim
                rank = nn.rank
                trust = nn.trust
                consensus = nn.consensus
                validator_trust = nn.validator_trust
                incentive = nn.incentive
                dividends = nn.dividends
                emission = int(nn.emission / (cosmos_tempo + 1) * 1e9)
                last_update = int(block - nn.last_update)
                validator_permit = nn.validator_permit
                row = [
                    hotwallet.name,
                    hotwallet.hotkey_str,
                    str(uid),
                    str(active),
                    "{:.5f}".format(stake),
                    "{:.5f}".format(rank),
                    "{:.5f}".format(trust),
                    "{:.5f}".format(consensus),
                    "{:.5f}".format(incentive),
                    "{:.5f}".format(dividends),
                    "{:_}".format(emission),
                    "{:.5f}".format(validator_trust),
                    "*" if validator_permit else "",
                    str(last_update),
                    nimble.utils.networking.int_to_ip(nn.fermion_info.ip)
                    + ":"
                    + str(nn.fermion_info.port)
                    if nn.fermion_info.port != 0
                    else "[yellow]none[/yellow]",
                    nn.hotkey,
                ]

                total_rank += rank
                total_trust += trust
                total_consensus += consensus
                total_incentive += incentive
                total_dividends += dividends
                total_emission += emission
                total_validator_trust += validator_trust

                if not (nn.hotkey, nn.coldkey) in hotkeys_seen:
                    # Don't double count stake on hotkey-coldkey pairs.
                    hotkeys_seen.add((nn.hotkey, nn.coldkey))
                    total_stake += stake

                # netuid -1 are particles that are de-registered.
                if netuid != "-1":
                    total_particles += 1

                TABLE_DATA.append(row)

            # Add cosmos header
            if netuid == "-1":
                grid.add_row(f"Deregistered Particles")
            else:
                grid.add_row(f"Cosmos: [bold white]{netuid}[/bold white]")

            table = Table(
                show_footer=False,
                width=cli.config.get("width", None),
                pad_edge=False,
                box=None,
            )
            if last_cosmos:
                table.add_column(
                    "[overline white]COLDKEY",
                    str(total_particles),
                    footer_style="overline white",
                    style="bold white",
                )
                table.add_column(
                    "[overline white]HOTKEY",
                    str(total_particles),
                    footer_style="overline white",
                    style="white",
                )
            else:
                # No footer for non-last cosmos.
                table.add_column("[overline white]COLDKEY", style="bold white")
                table.add_column("[overline white]HOTKEY", style="white")
            table.add_column(
                "[overline white]UID",
                str(total_particles),
                footer_style="overline white",
                style="yellow",
            )
            table.add_column(
                "[overline white]ACTIVE",
                justify="right",
                style="green",
                no_wrap=True,
            )
            if last_cosmos:
                table.add_column(
                    "[overline white]STAKE(nim)",
                    "nim{:.5f}".format(total_stake),
                    footer_style="overline white",
                    justify="right",
                    style="green",
                    no_wrap=True,
                )
            else:
                # No footer for non-last cosmos.
                table.add_column(
                    "[overline white]STAKE(nim)",
                    justify="right",
                    style="green",
                    no_wrap=True,
                )
            table.add_column(
                "[overline white]RANK",
                "{:.5f}".format(total_rank),
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]TRUST",
                "{:.5f}".format(total_trust),
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]CONSENSUS",
                "{:.5f}".format(total_consensus),
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]INCENTIVE",
                "{:.5f}".format(total_incentive),
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]DIVIDENDS",
                "{:.5f}".format(total_dividends),
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]EMISSION(\u03C1)",
                "\u03C1{:_}".format(total_emission),
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]VTRUST",
                "{:.5f}".format(total_validator_trust),
                footer_style="overline white",
                justify="right",
                style="green",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]VPERMIT", justify="right", no_wrap=True
            )
            table.add_column(
                "[overline white]UPDATED", justify="right", no_wrap=True
            )
            table.add_column(
                "[overline white]FERMION",
                justify="left",
                style="dim blue",
                no_wrap=True,
            )
            table.add_column(
                "[overline white]HOTKEY_SS58", style="dim blue", no_wrap=False
            )
            table.show_footer = True

            sort_by: Optional[str] = cli.config.get("sort_by", None)
            sort_order: Optional[str] = cli.config.get("sort_order", None)

            if sort_by is not None and sort_by != "":
                column_to_sort_by: int = 0
                highest_matching_ratio: int = 0
                sort_descending: bool = False  # Default sort_order to ascending

                for index, column in zip(
                    range(len(table.columns)), table.columns
                ):
                    # Fuzzy match the column name. Default to the first column.
                    column_name = column.header.lower().replace(
                        "[overline white]", ""
                    )
                    match_ratio = fuzz.ratio(sort_by.lower(), column_name)
                    # Finds the best matching column
                    if match_ratio > highest_matching_ratio:
                        highest_matching_ratio = match_ratio
                        column_to_sort_by = index

                if sort_order.lower() in {"desc", "descending", "reverse"}:
                    # Sort descending if the sort_order matches desc, descending, or reverse
                    sort_descending = True

                def overview_sort_function(row):
                    data = row[column_to_sort_by]
                    # Try to convert to number if possible
                    try:
                        data = float(data)
                    except ValueError:
                        pass
                    return data

                TABLE_DATA.sort(
                    key=overview_sort_function, reverse=sort_descending
                )

            for row in TABLE_DATA:
                table.add_row(*row)

            grid.add_row(table)

        console.clear()

        caption = "[italic][dim][white]Wallet balance: [green]nim" + str(
            total_balance.nim
        )
        grid.add_row(Align(caption, vertical="middle", align="center"))

        # Print the entire table/grid
        console.print(grid, width=cli.config.get("width", None))

    @staticmethod
    def _get_particles_for_netuid(
        args_tuple: Tuple["nimble.Config", int, List[str]]
    ) -> Tuple[int, List["nimble.ParticleInfoLite"], Optional[str]]:
        nbnetwork_config, netuid, hot_wallets = args_tuple

        result: List["nimble.ParticleInfoLite"] = []

        try:
            nbnetwork = nimble.nbnetwork(
                config=nbnetwork_config, log_verbose=False
            )

            all_particles: List["nimble.ParticleInfoLite"] = nbnetwork.particles_lite(
                netuid=netuid
            )
            # Map the hotkeys to uids
            hotkey_to_particles = {n.hotkey: n.uid for n in all_particles}
            for hot_wallet_addr in hot_wallets:
                uid = hotkey_to_particles.get(hot_wallet_addr)
                if uid is not None:
                    nn = all_particles[uid]
                    result.append(nn)
        except Exception as e:
            return netuid, [], "Error: {}".format(e)

        return netuid, result, None

    @staticmethod
    def _get_de_registered_stake_for_coldkey_wallet(
        args_tuple,
    ) -> Tuple[
        "nimble.Wallet", List[Tuple[str, "nimble.Balance"]], Optional[str]
    ]:
        nbnetwork_config, all_hotkey_addresses, coldkey_wallet = args_tuple

        # List of (hotkey_addr, our_stake) tuples.
        result: List[Tuple[str, "nimble.Balance"]] = []

        try:
            nbnetwork = nimble.nbnetwork(
                config=nbnetwork_config, log_verbose=False
            )

            # Pull all stake for our coldkey
            all_stake_info_for_coldkey = nbnetwork.get_stake_info_for_coldkey(
                coldkey_ss58=coldkey_wallet.coldkeypub.ss58_address
            )

            ## Filter out hotkeys that are in our wallets
            ## Filter out hotkeys that are delegates.
            def _filter_stake_info(stake_info: "nimble.StakeInfo") -> bool:
                if stake_info.stake == 0:
                    return False  # Skip hotkeys that we have no stake with.
                if stake_info.hotkey_ss58 in all_hotkey_addresses:
                    return False  # Skip hotkeys that are in our wallets.
                if nbnetwork.is_hotkey_delegate(
                    hotkey_ss58=stake_info.hotkey_ss58
                ):
                    return False  # Skip hotkeys that are delegates, they show up in nbcli my_delegates table.

                return True

            all_staked_hotkeys = filter(
                _filter_stake_info, all_stake_info_for_coldkey
            )
            result = [
                (stake_info.hotkey, stake_info.stake)
                for stake_info in all_staked_hotkeys
            ]

        except Exception as e:
            return coldkey_wallet, [], "Error: {}".format(e)

        return coldkey_wallet, result, None

    @staticmethod
    def add_args(parser: argparse.ArgumentParser):
        overview_parser = parser.add_parser(
            "overview", help="""Show registered account overview."""
        )
        overview_parser.add_argument(
            "--all",
            dest="all",
            action="store_true",
            help="""View overview for all wallets.""",
            default=False,
        )
        overview_parser.add_argument(
            "--width",
            dest="width",
            action="store",
            type=int,
            help="""Set the output width of the overview. Defaults to automatic width from terminal.""",
            default=None,
        )
        overview_parser.add_argument(
            "--sort_by",
            "--wallet.sort_by",
            dest="sort_by",
            required=False,
            action="store",
            default="",
            type=str,
            help="""Sort the hotkeys by the specified column title (e.g. name, uid, fermion).""",
        )
        overview_parser.add_argument(
            "--sort_order",
            "--wallet.sort_order",
            dest="sort_order",
            required=False,
            action="store",
            default="ascending",
            type=str,
            help="""Sort the hotkeys in the specified ordering. (ascending/asc or descending/desc/reverse)""",
        )
        overview_parser.add_argument(
            "--hotkeys",
            "--exclude_hotkeys",
            "--wallet.hotkeys",
            "--wallet.exclude_hotkeys",
            required=False,
            action="store",
            default=[],
            type=str,
            nargs="*",
            help="""Specify the hotkeys by name or ss58 address. (e.g. hk1 hk2 hk3)""",
        )
        overview_parser.add_argument(
            "--all_hotkeys",
            "--wallet.all_hotkeys",
            required=False,
            action="store_true",
            default=False,
            help="""To specify all hotkeys. Specifying hotkeys will exclude them from this all.""",
        )
        overview_parser.add_argument(
            "--netuid",
            dest="netuid",
            type=int,
            nargs="*",
            help="""Set the netuid(s) to filter by.""",
            default=[],
        )
        nimble.wallet.add_args(overview_parser)
        nimble.nbnetwork.add_args(overview_parser)

    @staticmethod
    def check_config(config: "nimble.config"):
        if (
            not config.is_set("wallet.name")
            and not config.no_predict
            and not config.get("all", d=None)
        ):
            wallet_name = Prompt.ask(
                "Enter wallet name", default=defaults.wallet.name
            )
            config.wallet.name = str(wallet_name)

        if config.netuid != []:
            if not isinstance(config.netuid, list):
                config.netuid = [int(config.netuid)]
            else:
                config.netuid = [int(netuid) for netuid in config.netuid]



"""
Implementation of the config class, which manages the config of different nimble modules.
"""
# The MIT License (MIT)
# Copyright © 2024 Nimble Labs Ltd

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of
# the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
import sys
import yaml
import copy
from copy import deepcopy
from munch import DefaultMunch
from typing import List, Optional, Dict, Any, TypeVar, Type
import argparse


class InvalidConfigFile(Exception):
    """In place of YAMLError"""

    pass


class config(DefaultMunch):
    """
    Implementation of the config class, which manages the config of different nimble modules.
    """

    __is_set: Dict[str, bool]

    r""" Translates the passed parser into a nested nimble config.
        Args:
            parser (argparse.ArgumentParser):
                Command line parser object.
            strict (bool):
                If true, the command line arguments are strictly parsed.
            args (list of str):
                Command line arguments.
            default (Optional[Any]):
                Default value for the Config. Defaults to None.
                This default will be returned for attributes that are undefined.
        Returns:
            config (nimble.config):
                Nested config object created from parser arguments.
    """

    def __init__(
        self,
        parser: argparse.ArgumentParser = None,
        args: Optional[List[str]] = None,
        strict: bool = False,
        default: Optional[Any] = None,
    ) -> None:
        super().__init__(default)

        self["__is_set"] = {}

        if parser == None:
            return None

        # Optionally add config specific arguments
        try:
            parser.add_argument(
                "--config",
                type=str,
                help="If set, defaults are overridden by passed file.",
            )
        except:
            # this can fail if --config has already been added.
            pass

        try:
            parser.add_argument(
                "--strict",
                action="store_true",
                help="""If flagged, config will check that only exact arguments have been set.""",
                default=False,
            )
        except:
            # this can fail if --strict has already been added.
            pass

        try:
            parser.add_argument(
                "--no_version_checking",
                action="store_true",
                help="Set true to stop cli version checking.",
                default=False,
            )
        except:
            # this can fail if --no_version_checking has already been added.
            pass

        try:
            parser.add_argument(
                "--no_predict",
                dest="no_predict",
                action="store_true",
                help="Set true to stop cli from inference the user.",
                default=False,
            )
        except:
            # this can fail if --no_version_checking has already been added.
            pass

        # Get args from argv if not passed in.
        if args == None:
            args = sys.argv[1:]

        # 1.1 Optionally load defaults if the --config is set.
        try:
            config_file_path = (
                str(os.getcwd())
                + "/"
                + vars(parser.parse_known_args(args)[0])["config"]
            )
        except Exception as e:
            config_file_path = None

        # Parse args not strict
        config_params = config.__parse_args__(
            args=args, parser=parser, strict=False
        )

        # 2. Optionally check for --strict
        ## strict=True when passed in OR when --strict is set
        strict = config_params.strict or strict

        if config_file_path != None:
            config_file_path = os.path.expanduser(config_file_path)
            try:
                with open(config_file_path) as f:
                    params_config = yaml.safe_load(f)
                    print(
                        "Loading config defaults from: {}".format(
                            config_file_path
                        )
                    )
                    parser.set_defaults(**params_config)
            except Exception as e:
                print(
                    "Error in loading: {} using default parser settings".format(
                        e
                    )
                )

        # 2. Continue with loading in params.
        params = config.__parse_args__(args=args, parser=parser, strict=strict)

        _config = self

        # Splits params and add to config
        config.__split_params__(params=params, _config=_config)

        # Make the is_set map
        _config["__is_set"] = {}

        ## Reparse args using default of unset
        parser_no_defaults = copy.deepcopy(parser)

        # Only command as the arg, else no args
        default_param_args = (
            [_config.get("command")]
            if _config.get("command") != None
            and _config.get("subcommand") == None
            else []
        )
        if _config.get("command") != None and _config.get("subcommand") != None:
            default_param_args = [
                _config.get("command"),
                _config.get("subcommand"),
            ]

        ## Get all args by name
        default_params = parser.parse_args(args=default_param_args)

        all_default_args = default_params.__dict__.keys() | []
        ## Make a dict with keys as args and values as argparse.SUPPRESS
        defaults_as_suppress = {
            key: argparse.SUPPRESS for key in all_default_args
        }
        ## Set the defaults to argparse.SUPPRESS, should remove them from the namespace
        parser_no_defaults.set_defaults(**defaults_as_suppress)
        parser_no_defaults._defaults.clear()  # Needed for quirk of argparse

        ### Check for subparsers and do the same
        if parser_no_defaults._subparsers != None:
            for action in parser_no_defaults._subparsers._actions:
                # Should only be the "command" subparser action
                if isinstance(action, argparse._SubParsersAction):
                    # Set the defaults to argparse.SUPPRESS, should remove them from the namespace
                    # Each choice is the keyword for a command, we need to set the defaults for each of these
                    ## Note: we also need to clear the _defaults dict for each, this is a quirk of argparse
                    cmd_parser: argparse.ArgumentParser
                    for cmd_parser in action.choices.values():
                        # If this choice is also a subparser, set defaults recursively
                        if cmd_parser._subparsers:
                            for action in cmd_parser._subparsers._actions:
                                # Should only be the "command" subparser action
                                if isinstance(
                                    action, argparse._SubParsersAction
                                ):
                                    cmd_parser: argparse.ArgumentParser
                                    for cmd_parser in action.choices.values():
                                        cmd_parser.set_defaults(
                                            **defaults_as_suppress
                                        )
                                        cmd_parser._defaults.clear()  # Needed for quirk of argparse
                        else:
                            cmd_parser.set_defaults(**defaults_as_suppress)
                            cmd_parser._defaults.clear()  # Needed for quirk of argparse

        ## Reparse the args, but this time with the defaults as argparse.SUPPRESS
        params_no_defaults = config.__parse_args__(
            args=args, parser=parser_no_defaults, strict=strict
        )

        ## Diff the params and params_no_defaults to get the is_set map
        _config["__is_set"] = {
            arg_key: True
            for arg_key in [
                k
                for k, _ in filter(
                    lambda kv: kv[1] != argparse.SUPPRESS,
                    params_no_defaults.__dict__.items(),
                )
            ]
        }

    @staticmethod
    def __split_params__(params: argparse.Namespace, _config: "config"):
        # Splits params on dot syntax i.e particle.fermion_port and adds to _config
        for arg_key, arg_val in params.__dict__.items():
            split_keys = arg_key.split(".")
            head = _config
            keys = split_keys
            while len(keys) > 1:
                if (
                    hasattr(head, keys[0]) and head[keys[0]] != None
                ):  # Needs to be Config
                    head = getattr(head, keys[0])
                    keys = keys[1:]
                else:
                    head[keys[0]] = config()
                    head = head[keys[0]]
                    keys = keys[1:]
            if len(keys) == 1:
                head[keys[0]] = arg_val

    @staticmethod
    def __parse_args__(
        args: List[str],
        parser: argparse.ArgumentParser = None,
        strict: bool = False,
    ) -> argparse.Namespace:
        """Parses the passed args use the passed parser.
        Args:
            args (List[str]):
                List of arguments to parse.
            parser (argparse.ArgumentParser):
                Command line parser object.
            strict (bool):
                If true, the command line arguments are strictly parsed.
        Returns:
            Namespace:
                Namespace object created from parser arguments.
        """
        if not strict:
            params, unrecognized = parser.parse_known_args(args=args)
            params_list = list(params.__dict__)
            # bug within argparse itself, does not correctly set value for boolean flags
            for unrec in unrecognized:
                if unrec.startswith("--") and unrec[2:] in params_list:
                    # Set the missing boolean value to true
                    setattr(params, unrec[2:], True)
        else:
            params = parser.parse_args(args=args)

        return params

    def __deepcopy__(self, memo) -> "config":
        _default = self.__default__

        config_state = self.__getstate__()
        config_copy = config()
        memo[id(self)] = config_copy

        config_copy.__setstate__(config_state)
        config_copy.__default__ = _default

        config_copy["__is_set"] = deepcopy(self["__is_set"], memo)

        return config_copy

    def __repr__(self) -> str:
        return self.__str__()

    @staticmethod
    def _remove_private_keys(d):
        if "__parser" in d:
            d.pop("__parser", None)
        if "__is_set" in d:
            d.pop("__is_set", None)
        for k, v in list(d.items()):
            if isinstance(v, dict):
                config._remove_private_keys(v)
        return d

    def __str__(self) -> str:
        # remove the parser and is_set map from the visible config
        visible = copy.deepcopy(self.toDict())
        visible.pop("__parser", None)
        visible.pop("__is_set", None)
        cleaned = config._remove_private_keys(visible)
        return "\n" + yaml.dump(cleaned, sort_keys=False)

    def copy(self) -> "config":
        return copy.deepcopy(self)

    def to_string(self, items) -> str:
        """Get string from items"""
        return "\n" + yaml.dump(items.toDict())

    def update_with_kwargs(self, kwargs):
        """Add config to self"""
        for key, val in kwargs.items():
            self[key] = val

    @classmethod
    def _merge(cls, a, b):
        """Merge two configurations recursively.
        If there is a conflict, the value from the second configuration will take precedence.
        """
        for key in b:
            if key in a:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    a[key] = cls._merge(a[key], b[key])
                else:
                    a[key] = b[key]
            else:
                a[key] = b[key]
        return a

    def merge(self, b):
        """
        Merges the current config with another config.

        Args:
            b: Another config to merge.
        """
        self = self._merge(self, b)

    @classmethod
    def merge_all(cls, configs: List["config"]) -> "config":
        """
        Merge all configs in the list into one config.
        If there is a conflict, the value from the last configuration in the list will take precedence.

        Args:
            configs (list of config):
                List of configs to be merged.

        Returns:
            config:
                Merged config object.
        """
        result = cls()
        for cfg in configs:
            result.merge(cfg)
        return result

    def is_set(self, param_name: str) -> bool:
        """
        Returns a boolean indicating whether the parameter has been set or is still the default.
        """
        if param_name not in self.get("__is_set"):
            return False
        else:
            return self.get("__is_set")[param_name]


T = TypeVar("T", bound="DefaultConfig")


class DefaultConfig(config):
    """
    A Config with a set of default values.
    """

    @classmethod
    def default(cls: Type[T]) -> T:
        """
        Get default config.
        """
        raise NotImplementedError("Function default is not implemented.")
